"""
links.py
This module defines the Link types for ActivityPub.
These links are based on the ActivityStreams vocabulary.

For more information, see:
https://www.w3.org/TR/activitystreams-vocabulary/#link
"""

from typing import Any
from pydantic import BaseModel, Field, field_validator

from phederation.utils.serialization import ActivityPubBase


class RelativeLink(BaseModel):
    """Relative link model to indicate where actor information can be found."""

    rel: str = Field(
        ..., description="Relation of this link template."
    )
    type: None | str = Field(
        default='application/json',
        description="The media type, i.e. 'application/activity+json' or 'text/html'.",
    )
    href: str


class RelativeLinkTemplate(BaseModel):
    rel: str = Field(
        ..., description="Relation of this link template."
    )
    template: None | str = Field(
        default='<instance_url>/authorize_interaction?uri={{uri}}',
        description="How to access the information in the link.",
    )

class APLink(ActivityPubBase):
    """
    Represents a Link in ActivityPub.
    
    A Link is an indirect, qualified reference to a resource identified by a URL.
    It can be used to represent a variety of connections between objects in the ActivityPub ecosystem.

    Specification: https://www.w3.org/TR/activitystreams-vocabulary/#link
    Usage: https://www.w3.org/TR/activitypub/#object-without-create
    """
    type: str = Field(default="Link", description="Indicates that this object represents a link")
    href: str = Field(..., description="The target resource pointed to by the link")
    rel: None | list[str] = Field(default=None, description="The relationship between the resource and the link")
    media_type: None | str = Field(default=None, description="The MIME media type of the referenced resource")
    name: None | str = Field(default=None, description="A human-readable name for the link")
    hreflang: None | str = Field(default=None, description="The language of the referenced resource")
    height: None | int = Field(default=None, description="The height of the referenced resource in pixels")
    width: None | int = Field(default=None, description="The width of the referenced resource in pixels")
    preview: None | str = Field(default=None, description="A link to a preview of the referenced resource")

    @field_validator('media_type')
    def validate_media_type(cls, value: Any | None):
        allowed_mimes = ["image/jpeg", "image/png", "application/json", "text/html", "video/mp4"]
        if value and not value in allowed_mimes:
            raise ValueError(f"Invalid MIME type: {value}")
        return value

class APMention(APLink):
    """
    Represents a Mention in ActivityPub.
    
    A Mention is a specialized Link that represents an @mention.
    It is typically used to reference another actor in the content of an object.

    Properties:
        type (Literal["Mention"]): The type of the object, always "Mention" for this class.
        href (ObjectId): The URL of the mentioned actor.
        name (None | str): The name or username of the mentioned actor.

    Inherits all other properties from APLink.

    Specification: https://www.w3.org/TR/activitystreams-vocabulary/#mention
    Usage: https://www.w3.org/TR/activitypub/#mention
    """
    type: str = Field(default="Mention", description="Indicates that this object represents a mention")



"""
activities.py
This module defines the Activity types for ActivityPub.
These activities are based on the ActivityStreams vocabulary.

For more information, see:
https://www.w3.org/TR/activitystreams-vocabulary/#activity-types
"""

from typing import ClassVar, override

from pydantic import ConfigDict, Field, SerializeAsAny

from phederation.utils import ActivityPubBaseWithId, ActivityType, ObjectId
from phederation.utils.base import APDateTime

from .errors import BaseError
from .objects import APDocument, APObject


class ErrorMessageActivityNotFound(BaseError):
    message: str = "Activity does not exist locally."
    description: str = "The activity with the given id cannot be found on this instance."


class ErrorMessageActivityRestricted(BaseError):
    message: str = "Activity has restricted visibility."
    description: str = "The activity with the given id has restricted visibility."


class ErrorMessageActivityValidation(BaseError):
    message: str = "Activity request body is not valid."
    description: str = "The activity posted has an invalid request body."


class ErrorMessageActivityHandler(BaseError):
    message: str = "Activity could not be processed properly."
    description: str = "The inbox could not process the given activity request body."


class APActivity(APObject):
    """
    Base class for all ActivityPub activities.
    An Activity is a subtype of Object that describes some form of action.

    Specification: https://www.w3.org/TR/activitystreams-vocabulary/#activity
    Usage: https://www.w3.org/TR/activitypub/#activities
    """

    actor: str | SerializeAsAny[APObject] = Field(..., description="The actor performing the activity")
    object: None | str | SerializeAsAny[APObject] = Field(default=None, description="The object of the activity")
    target: None | str | SerializeAsAny[APObject] = Field(default=None, description="The target of the activity")
    result: None | str | SerializeAsAny[APObject] = Field(default=None, description="The result of the activity")
    origin: None | str | SerializeAsAny[APObject] = Field(default=None, description="The origin of the activity")
    instrument: None | str | SerializeAsAny[APObject] = Field(default=None, description="The instrument used")

    @override
    def clear_recipients(self):
        super().clear_recipients()
        if self.object and not isinstance(self.object, str):
            self.object.clear_recipients()

    @override
    async def get_recipients(self) -> list[ObjectId]:
        """Get the (deduplicated) list of actors in the activity recipient fields (to,cc,bto,bcc,audience).
        This includes the activity.object recipients. Does *not* resolve the inboxes of these recipients yet.

        Returns:
            list[ObjectId]: List of actor_id's for all recipients found.
        """
        recipients = await super().get_recipients()
        object = self.object
        if isinstance(object, dict):
            object = APObject.deserialize(object)
        if isinstance(object, APObject):
            recipients += await object.get_recipients()

        recipients = list(set(recipients))
        return recipients

    def get_object_or_None(self):
        """Returns the activity.object as APObject or APDocument if it is set (also if it is a dict).

        If the activity.object is None or just an ObjectId string, this returns None.

        Returns:
            APObject | None: the object or None.
        """
        if isinstance(self.object, APObject):
            return self.object

    model_config: ClassVar[ConfigDict] = ConfigDict(
        json_schema_extra={
            "examples": [
                {
                    "id": "https://example.org/id-of-the-activity-unused",
                    "type": "Create",
                    "attributedTo": "https://example.org/id-of-an-actor",
                    "content": "The content of the activity described in text.",
                    "summary": "A summary of the activity.",
                    "actor": "https://example.org/id-of-actor-performing-the-activity",
                    "object": {"id": "https://example.org/id-of-the-note-temporary", "type": "Note", "summary": "This is a note to create."},
                    "target": "https://example.org/id-of-target-actor-or-object",
                }
            ]
        }
    )


class APCreate(APActivity):
    """
    Represents a Create activity in ActivityPub.
    Indicates that the actor has created the object.

    Specification: https://www.w3.org/TR/activitystreams-vocabulary/#create
    """

    type: str = Field(
        default=ActivityType.CREATE.value,
        description="Indicates that this object represents a create activity",
    )
    object: None | str | APDocument | APObject = Field(default=None, description="The object being created")


class APUpdate(APActivity):
    """
    Represents an Update activity in ActivityPub.
    Indicates that the actor has updated the object.

    Specification: https://www.w3.org/TR/activitystreams-vocabulary/#update
    """

    type: str = Field(
        default=ActivityType.UPDATE.value,
        description="Indicates that this object represents an update activity",
    )
    object: None | str | APObject = Field(default=None, description="The object being updated")


class APDelete(APActivity):
    """
    Represents a Delete activity in ActivityPub.
    Indicates that the actor has deleted the object.

    Specification: https://www.w3.org/TR/activitystreams-vocabulary/#delete
    """

    type: str = Field(
        default=ActivityType.DELETE.value,
        description="Indicates that this object represents a delete activity",
    )
    object: None | str | APObject = Field(default=None, description="The object being deleted")


class APFollow(APActivity):
    """
    Represents a Follow activity in ActivityPub.
    Indicates that the actor is "following" the object.

    Specification: https://www.w3.org/TR/activitystreams-vocabulary/#follow
    """

    type: str = Field(
        default=ActivityType.FOLLOW.value,
        description="Indicates that this object represents a follow activity",
    )
    object: None | str | APObject = Field(default=..., description="The object being followed")
    accepted: bool = Field(default=False, description="If the object has accepted the follow of actor")
    accepted_at: APDateTime | None = Field(default=None, description="Time of accept of follow.")


class APUndo(APActivity):
    """
    Represents an Undo activity in ActivityPub.
    Indicates that the actor is undoing the object activity.

    Specification: https://www.w3.org/TR/activitystreams-vocabulary/#undo
    """

    type: str = Field(
        default=ActivityType.UNDO.value,
        description="Indicates that this object represents an undo activity",
    )
    object: None | str | APObject = Field(default=None, description="The activity being undone")


class APLike(APActivity):
    """
    Represents a Like activity in ActivityPub.
    Indicates that the actor likes the object.

    Specification: https://www.w3.org/TR/activitystreams-vocabulary/#like
    """

    type: str = Field(
        default=ActivityType.LIKE.value,
        description="Indicates that this object represents a like activity",
    )
    object: None | str | APObject = Field(default=None, description="The object being liked")


class APAnnounce(APActivity):
    """
    Represents an Announce activity in ActivityPub.
    Indicates that the actor is calling the target's attention to the object.

    Specification: https://www.w3.org/TR/activitystreams-vocabulary/#announce
    """

    type: str = Field(
        default=ActivityType.ANNOUNCE.value,
        description="Indicates that this object represents an announce activity",
    )
    object: None | str | APObject = Field(default=None, description="The object being announced")


class APAccept(APActivity):
    """
    Represents an Accept activity in ActivityPub.
    Indicates that the actor accepts the object.

    Specification: https://www.w3.org/TR/activitystreams-vocabulary/#accept
    """

    type: str = Field(
        default=ActivityType.ACCEPT.value,
        description="Indicates that this object represents an accept activity",
    )
    object: None | str | APObject = Field(default=None, description="The object being accepted")


class APAdd(APActivity):
    """
    Represents a Add activity in ActivityPub.
    Indicates that the actor is adding the object.

    Specification: https://www.w3.org/TR/activitystreams-vocabulary/#add
    """

    type: str = Field(
        default=ActivityType.ADD.value,
        description="Activity to add an object to an origin",
    )
    object: None | str | APObject = Field(default=None, description="The object being added")
    origin: None | str | APObject = Field(default=None, description="The origin where the object should be added to")


class APRemove(APActivity):
    """
    Represents a Remove activity in ActivityPub.
    Indicates that the actor is removing the object.

    Specification: https://www.w3.org/TR/activitystreams-vocabulary/#remove
    """

    type: str = Field(
        default=ActivityType.REMOVE.value,
        description="Activity to remove an object from an origin",
    )
    object: None | str | APObject = Field(default=None, description="The object being removed")
    origin: None | str | APObject = Field(default=None, description="The origin where the object resides now, and where it should be removed from")


class APBlock(APActivity):
    """
    Represents a Block activity in ActivityPub.
    Indicates that the actor is blocking the object.

    Specification: https://www.w3.org/TR/activitystreams-vocabulary/#block
    """

    type: str = Field(
        default=ActivityType.BLOCK.value,
        description="Indicates that this object represents a block activity",
    )
    object: None | str | APObject = Field(default=None, description="The object being blocked")


class APReject(APActivity):
    """
    Represents a Reject activity in ActivityPub.
    Indicates that the actor rejects the object.
    This is typically used to reject incoming activities such as Follow requests.

    Specification: https://www.w3.org/TR/activitystreams-vocabulary/#reject
    """

    type: str = Field(
        default=ActivityType.REJECT.value,
        description="A reject activity",
    )
    object: None | str | APObject = Field(default=None, description="The object being rejected")


class APMigrate(APActivity):
    """Represents a Migrate activity.

    This is not part of the official ActivityPub specification.
    It allows to migrate user accounts to new instances, or between accounts on the same instance.
    """

    type: str = Field(default=ActivityType.MIGRATE.value, description="A migrate activity")
    actor_to: ObjectId | None = Field(
        default=None,
        description="Actor to migrate to. Can be on a new instance, or on the same instance. If set to None, this can be used to download the migration data locally and use it for migration somewhere else",
    )
    accepted_at: APDateTime | None = Field(
        default=None,
        description="If the actor on the new instance has accepted the APMigrate activity. This is the last step before migration is completed",
    )


class ActivityInQueue(ActivityPubBaseWithId):
    type: str = ActivityType.ACTIVITY_IN_QUEUE.value
    activity: APActivity = Field(..., description="The activity to send")
    recipients: list[ObjectId] = Field(
        ...,
        description="The list of ids of the recipient actors. This can include the bcc/bto fields that should be removed already from the given activity",
    )
    send_at: APDateTime = Field(..., description="When to attempt to send the given activity")



from .account import AccountRoute
from .activity import ActivityRoute
from .actor import ActorRoute
from .instance import InstanceRoute
from .key import KeyRoute
from .media import MediaRoute
from .object import ObjectRoute
from .proof import ProofRoute
from .security import SecurityRoute
from .webfinger import WebfingerRoute
from .admin_instance import AdminInstanceRoute

__all__ = [
    "ActorRoute",
    "AccountRoute",
    "ActivityRoute",
    "WebfingerRoute",
    "ObjectRoute",
    "KeyRoute",
    "ProofRoute",
    "InstanceRoute",
    "SecurityRoute",
    "MediaRoute",
    "AdminInstanceRoute"
]



"""media.py

The media route, to serve and accept media objects.
"""

from http import HTTPStatus
import json
from typing import Annotated, Any, cast, override

from fastapi import Depends, File, HTTPException
from fastapi.datastructures import UploadFile
from fastapi.encoders import jsonable_encoder
from fastapi.params import Form, Security
from fastapi.responses import JSONResponse, StreamingResponse

from phederation.api.routes.base import BaseRoute
from phederation.models import APActivity, ErrorMessageActivityHandler
from phederation.models import (
    ErrorMessageActivityRestricted,
)
from phederation.models.objects import APDocument
from phederation.models.activities import ErrorMessageActivityValidation
from phederation.models.actors import APActor, ErrorMessageActorUnauthorized
from phederation.security.authentication import UserInfo
from phederation.utils.base import UrlType, assemble_id_url
from phederation.utils.exceptions import (
    ActivityPubException,
    HandlerError,
    SecurityError,
    UserError,
)


class MediaRoute(BaseRoute):

    @override
    def setup(self):

        @self.app.get("/media/{media_primary}", tags=["media"], response_description="A stream to download the media object (file), if it exists")
        async def get_media(  # pyright: ignore[reportUnusedFunction]
            media_primary: str, user_info: Annotated[UserInfo | None, Security(self.middleware.authentication, scopes=["user"])] = None
        ) -> StreamingResponse:
            """Return media with given id, if accessible."""
            if "media" not in self.api.settings.federation.endpoints:
                raise HTTPException(status_code=HTTPStatus.METHOD_NOT_ALLOWED, detail="Instance does not support the media endpoint")

            file_id = assemble_id_url(
                base_url=self.api.settings.domain.base_url,
                primary=media_primary,
                type=UrlType.Media,
            )
            metadata = await self.server.storage.metadata(file_id=file_id)
            document_id = metadata.document_id

            actor_requesting_access = user_info.actor_id if user_info else None

            document = await self.api.server.resolver.resolve_object(object_id=document_id)
            obj_with_access = await self.api.server.resolver.resolve_access_type(object=document, actor_requesting_access=actor_requesting_access)
            if not obj_with_access or not isinstance(obj_with_access, APDocument):
                self.logger.warning(f"Media file visibility restricted, user: {user_info} cannot access it")
                raise HTTPException(
                    status_code=HTTPStatus.BAD_REQUEST,
                    detail=ErrorMessageActivityRestricted().message,
                )

            # load the actual file from storage and stream it to the client
            obj = await self.server.storage.load(file_id=file_id)
            return StreamingResponse(obj)

        @self.router.post(
            "/users/{username}/endpoints/uploadMedia",
            tags=["users"],
            responses={
                HTTPStatus.BAD_REQUEST: {
                    "model": ErrorMessageActivityValidation,
                    "description": ErrorMessageActivityValidation().description,
                },
                HTTPStatus.INTERNAL_SERVER_ERROR: {
                    "model": ErrorMessageActivityHandler,
                    "description": ErrorMessageActivityHandler().description,
                },
                HTTPStatus.UNAUTHORIZED: {
                    "model": ErrorMessageActorUnauthorized,
                    "description": ErrorMessageActorUnauthorized().description,
                },
            },
            dependencies=[Security(self.middleware.authorization, scopes=["user"])],
        )
        async def post_user_endpoints(  # pyright: ignore[reportUnusedFunction]
            actor: Annotated[APActor, Depends(self.middleware.authentication.verify_local_username)],
            activity_form: Annotated[str, Form(..., description="Create activity sent in addition of the file data")],
            file: Annotated[UploadFile, File(..., description="File to upload to the media endpoint")],
        ):
            """Handle user media endpoint requests."""
            if "media" not in self.api.settings.federation.endpoints:
                raise HTTPException(status_code=HTTPStatus.METHOD_NOT_ALLOWED, detail="Instance does not support the media endpoint")

            if actor.id is None:
                raise UserError("Actor does not have an id")

            activity_dict = cast(dict[str, Any], json.loads(activity_form))
            activity = self.api.server.resolver.resolve_from_dict(activity_dict)
            if not isinstance(activity, APActivity):
                raise HandlerError("Object sent was not an APActivity")
            activity.attributed_to = actor.id
            activity = await self.server.media_handler.validate(activity=activity, actor_id=actor.id)

            # we can now create the media file and store it in the activity.
            if self.api.settings.media.create_file_hash:
                actor_current_key = await self.server.key_manager.resolve_current_key(actor.public_key_as_ids())
                if not actor_current_key:
                    raise SecurityError("Actor public key not found")

                async def proof_string(string_value: str):
                    return await self.server.key_manager.proof_string(string_to_sign=string_value, private_key_id=actor_current_key.id)

                document = await self.server.media_handler.process_attachment(attachment=file, description=activity.object, proof_string=proof_string)

                document.attributed_to = actor.id
                activity.object = document

            result = "Result unset"
            try:
                result = await self.server.handle_outbox(activity=activity)
                activity_id = result.get_success_id()
                if not activity_id:
                    raise HandlerError(f"Did not create activity id in outbox, result={result}")
            except ActivityPubException as e:
                self.logger.error(f"Exception in post_user_endpoints: {e.message}")
                raise HTTPException(detail=e.user_message, status_code=e.status_code)

            # Servers MUST return a 201 Created HTTP code, and unless the activity is transient, MUST include the new id in the Location header.
            # https://www.w3.org/TR/activitypub/#client-to-server-interactions
            return JSONResponse(
                content=jsonable_encoder({"Status": "OK"}),
                status_code=HTTPStatus.CREATED,
                media_type="application/activity+json",
                headers={"Location": activity_id},
            )



"""
media.py
This module defines metadata for media sent/received through activity pub.

For more information, see:
https://www.w3.org/TR/activitystreams-vocabulary/
"""

from pydantic import BaseModel, Field
from phederation.utils.base import ObjectId

class MediaMetadata(BaseModel):
    """
    Represents media for a file object sent over the media endpoint of APActors.
    """
    filename: str = Field(default=..., description='Name of the file. Can be as it is stored on the server, or as it was defined by the uploading actor.')
    content_type: str = Field(..., description="Mime type of the file.")
    hash_sha256: str | None = Field(default=None, description="Sha256 hash of the file content. Optional, used by DataIntegrityProof process.")
    document_id: ObjectId = Field(default=..., description="Id of the APDocument that describes this file, including the accessibility.")


from datetime import datetime, timedelta, timezone
from typing import Any, override

from phederation.cache.base import BaseCache


class DictCache(BaseCache):
    def __init__(self, max_size: int = 100, ttl: int = 300):
        self.max_size: int = max_size
        self.ttl: int = ttl
        self._cache: dict[str, tuple[Any | None, datetime]] = {}

    @override
    def get(self, key: str) -> None | Any:
        if key in self._cache:
            value, timestamp = self._cache[key]
            if datetime.now(timezone.utc) - timestamp < timedelta(seconds=self.ttl):
                return value
            else:
                del self._cache[key]
        return None

    @override
    def set(self, key: str, value: Any | None):
        if len(self._cache) >= self.max_size:
            oldest_key = min(self._cache, key=lambda k: self._cache[k][1])
            del self._cache[oldest_key]
        self._cache[key] = (value, datetime.now(timezone.utc))

    @override
    def delete(self, key: str):
        _ = self._cache.pop(key, None)

    @override
    def clear(self, prefix: str | None = None):
        if not prefix:
            self._cache.clear()
        else:
            keys = list(self._cache.keys())
            for key in keys:
                if key.startswith(prefix):
                    _ = self._cache.pop(key)

    @override
    def close(self):
        self.clear()



from http import HTTPStatus
from typing import Annotated, override

from fastapi import Path

from phederation.api.routes.base import BaseRoute
from phederation.models.errors import BaseError
from phederation.models.proofs import DataIntegrityProof
from phederation.utils.base import UrlType, assemble_id_url
from phederation.utils.exceptions import SecurityError


class ErrorMessageProofNotFound(BaseError):
    message: str = "Proof not found"


class ProofRoute(BaseRoute):

    @override
    def setup(self):

        @self.router.get(
            "/proofs/{proof_primary}",
            status_code=HTTPStatus.OK,
            response_model_exclude_none=True,
            responses={
                HTTPStatus.NOT_FOUND: {
                    "model": ErrorMessageProofNotFound,
                    "description": "If the proof with the given primary cannot be found on this instance.",
                }
            },
            tags=["proofs"],
        )
        async def get_proof(  # pyright: ignore[reportUnusedFunction]
            proof_primary: Annotated[str, Path(description="The identifier of the key (without the URL) on this instance.")],
        ) -> DataIntegrityProof:
            """Handle requests for proofs.

            Visibility is not checked here, all proofs are public.
            """
            self.logger.debug(f"Received request for proof primary: {proof_primary}")

            # Return key document with proper JSON-LD context
            proof_id = assemble_id_url(type=UrlType.Proofs, base_url=self.api.settings.domain.base_url, primary=proof_primary)
            proof = await self.api.server.storage.proof.read(id=proof_id)
            if not proof:
                log_message = f"No proof found with primary '{proof_primary}'"
                raise SecurityError(message=log_message, user_facing_message=ErrorMessageProofNotFound().message, status_code=HTTPStatus.NOT_FOUND)

            return proof



"""
Collection handling implementation.
"""

from abc import ABC, abstractmethod
from collections.abc import AsyncGenerator
from datetime import datetime, timezone
from logging import Logger
from typing import Any
from urllib.parse import parse_qsl, urlencode, urlparse, urlunparse

from phederation.models import APOrderedCollectionPage
from phederation.storage.base import StorageBackend
from phederation.utils import ObjectId
from phederation.utils.base import AccessType, urljoin
from phederation.utils.exceptions import CollectionError, catch_exceptions
from phederation.utils.logging import configure_logger


class CollectionManager(ABC):
    """Manages ActivityPub collections."""

    def __init__(self, storage: StorageBackend, page_size: int = 20, access_type: AccessType = AccessType.PUBLIC):
        self.storage: StorageBackend = storage
        self.page_size: int = page_size
        self.logger: Logger = configure_logger(__name__, prefix=self.storage.settings.federation.logging_prefix)
        self.access_type: AccessType = access_type

    @staticmethod
    def collection_id(id: ObjectId, name: str, page: int | None = None) -> ObjectId:
        """Creates a resolvable id for a collection with given name."""
        url = urljoin(id, name)
        if not page:
            return url
        else:
            return f"{url}?page={page}"

    @staticmethod
    def collection_page(id: ObjectId, page: int | None = None) -> ObjectId:
        """Creates a resolvable id for a collection page with given collection id."""
        if page is None:
            return id
        else:
            # set the page element of the URL to the given integer
            url_parts = list(urlparse(id))
            query = dict(parse_qsl(url_parts[4]))
            query.update({"page": str(page)})
            url_parts[4] = urlencode(query)
            return urlunparse(url_parts)

    @abstractmethod
    async def initialize(self):
        pass

    async def create_new_page(
        self,
        collection_id: ObjectId,
        page: int,
        n_pages: int,
        total_items: int,
        attributed_to: ObjectId | None = None,
        items: list[ObjectId] | None = None,
    ) -> APOrderedCollectionPage:
        if items is None or len(items) == 0:
            items = None
        return APOrderedCollectionPage(
            id=CollectionManager.collection_page(id=collection_id, page=page),
            attributed_to=attributed_to,
            total_items=total_items,
            part_of=collection_id,
            ordered_items=items,
            published=datetime.now(timezone.utc),
            start_index=1,
            current=CollectionManager.collection_page(id=collection_id, page=1),
            first=CollectionManager.collection_page(id=collection_id, page=1),
            prev=(CollectionManager.collection_page(id=collection_id, page=page - 1) if page - 1 > 0 else None),
            next=(CollectionManager.collection_page(id=collection_id, page=page + 1) if page + 1 <= n_pages else None),
            last=CollectionManager.collection_page(id=collection_id, page=n_pages),
        )

    @catch_exceptions(CollectionError, "Failed in create_collection")
    @abstractmethod
    async def create_collection(
        self,
        collection_id: ObjectId,
        items: list[ObjectId] | None = None,
        attributed_to: ObjectId | None = None,
        access: AccessType = AccessType.PUBLIC,
    ) -> ObjectId:
        """
        Create a new collection.

        Args:
            collection_id: Collection id to create
            items: Initial collection items
            attributed_to: the actor or object this collection is attributed to. This actor always has access to all items in the collection.

        Returns:
            Collection ID
        """
        pass

    @abstractmethod
    def pages(
        self, collection_id: ObjectId, attributed_to: ObjectId | None = None, access: AccessType = AccessType.PUBLIC
    ) -> AsyncGenerator[APOrderedCollectionPage, Any]:
        """Iterate over all pages of the collection.

        Args:
            collection_id (ObjectId): id of the collection. Can include a page, but will always start with the first item.

        Raises:
            CollectionError: If a part of the collection cannot be found.

        Yields:
            APOrderedCollectionPage: pages of the collection, one by one.
        """
        pass

    @abstractmethod
    async def contains(self, collection_id: ObjectId, item: ObjectId | None, access: AccessType = AccessType.PUBLIC) -> bool:
        """Check if the collection contains the given item.

        Args:
            collection_id (ObjectId): id of the collection
            item (ObjectId | None): item to check. Always returns False if item is None.

        Returns:
            bool: True if contained, False otherwise.
        """
        pass

    @catch_exceptions(CollectionError, "Failed in add_to_collection")
    @abstractmethod
    async def add_to_collection(
        self, collection_id: ObjectId, items: ObjectId | list[ObjectId], access: AccessType = AccessType.PUBLIC, add_only_once: bool = False
    ) -> None:
        """
        Add items to collection.

        Args:
            collection_id: Collection ID
            items: Item(s): to add
            access: AccessType: the accessibility level of the items to add. This can be independent of the access type of the collection, but not less restrictive.
                               If the access type of the colleciton is PRIVATE, adding PUBLIC elements will not make them accessible.
            add_only_once (bool): if True, will not add any item id more than once. Default: False.
        """
        pass

    @abstractmethod
    async def remove_from_collection(self, collection_id: ObjectId, items: ObjectId | list[ObjectId], access: AccessType = AccessType.PUBLIC) -> None:
        """
        Remove items from collection.
        This deletes the entire collection, including all pages, and reforms it - meaning it is not very efficient.

        Args:
            collection_id: Collection ID
            items: Item(s) to remove
            access: AccessType: the access type used for removal. Only elements with accesss type like this or more accessible can be removed.
        """
        pass

    @abstractmethod
    async def delete_collection(self, collection_id: ObjectId):
        """Deletes the entire collection, including all pages, from storage.
        Items within the collection are not affected.

        Args:
            collection_id (ObjectId): The collection to remove.
        """
        pass

    @abstractmethod
    async def collection_exists(self, collection_id: ObjectId, page: int | None = None, access: AccessType = AccessType.PUBLIC) -> bool:
        """Check if the collection exists.

        Args:
            collection_id (ObjectId): id of the collection to check.
            page (int | None, optional): Optional, only check if the given page exists. Defaults to None.

        Returns:
            APOrderedCollectionPage | None: The collection if it exists, or None.
        """
        pass

    @catch_exceptions(CollectionError, "Failed in get_collection_page")
    @abstractmethod
    async def get_collection_page(
        self, collection_id: ObjectId, attributed_to: ObjectId | None = None, page: int = 1, access: AccessType = AccessType.PUBLIC
    ) -> APOrderedCollectionPage:
        """
        Get collection page.

        Args:
            collection_id: Collection ID
            page: Page number

        Returns:
            Collection page
        """
        pass

    @abstractmethod
    async def get_collection_items(self, collection_id: ObjectId, access: AccessType = AccessType.PUBLIC) -> list[ObjectId]:
        """Returns all item ids of all pages in the entire collection, not paginated.

        Args:
            collection_id (ObjectId): Id of the collection
            access (AccessType): only retrieve items with this access type (or more permissible). Default: PUBLIC.

        Returns:
            list[ObjectId]: All item ids.
        """
        pass



"""Utilities used in other modules."""

from .base import (
    ActivityPubBaseWithId,
    RateLimit,
    ActivityType,
    ObjectId,
    APDateTime,
    ObjectType,
    CollectionType,
    PUBLIC_URLS,
    UrlType,
    actor_id_from_username,
    assemble_id_url,
    collection_id_from_name,
    validate_object_id,
    ResolverResult,
    NodeInfo,
)
from .exceptions import *
from .logging import configure_logger
from .settings import (
    DatabaseSettings,
    DomainSettings,
    FederationSettings,
    KeycloakSettings,
    MediaSettings,
    PhedSettings,
    SecuritySettings,
    StatisticsSettings,
    StorageSettings,
)
from .version import PHEDERATION_SOFTWARE_NAME, PHEDERATION_USERAGENT, PHEDERATION_VERSION, PHEDERATION_HEADERS

__all__ = [
    "NodeInfo",
    "RateLimit",
    "ResolverResult",
    "configure_logger",
    "DatabaseSettings",
    "DomainSettings",
    "FederationSettings",
    "KeycloakSettings",
    "MediaSettings",
    "PhedSettings",
    "SecuritySettings",
    "StatisticsSettings",
    "StorageSettings",
    "ObjectId",
    "APDateTime",
    "ActivityPubBaseWithId",
    "PUBLIC_URLS",
    "UrlType",
    "ActivityType",
    "ObjectType",
    "CollectionType",
    "assemble_id_url",
    "validate_object_id",
    "actor_id_from_username",
    "collection_id_from_name",
    "PHEDERATION_VERSION",
    "PHEDERATION_SOFTWARE_NAME",
    "PHEDERATION_USERAGENT",
    "PHEDERATION_HEADERS",
]



"""
Enhanced Create activity handler.
"""

from typing import cast, override

from phederation.models import dereference
from phederation.models.activities import APActivity, APCreate
from phederation.models.keys import APPublicKey
from phederation.models.objects import (
    APArticle,
    APDocument,
    APImage,
    APNote,
    APObject,
    APUserData,
    DocumentType,
)
from phederation.models.proofs import DataIntegrityProof
from phederation.utils import ObjectId
from phederation.utils.base import (
    AccessType,
    ActivityType,
    ObjectType,
    collection_id_from_name,
)
from phederation.utils.exceptions import HandlerError, SecurityError, ValidationError

from .base import ActivityHandler


class CreateHandler(ActivityHandler):
    """Enhanced Create activity handler."""

    SUPPORTED_TYPES: dict[str, type[APNote] | type[APArticle] | type[APImage] | type[APDocument] | type[APUserData] | type[APPublicKey]] = {
        DocumentType.NOTE.value: APNote,
        DocumentType.ARTICLE.value: APArticle,
        DocumentType.IMAGE.value: APImage,
        DocumentType.DOCUMENT.value: APDocument,
        ObjectType.KEY.value: APPublicKey,
    }

    @classmethod
    def wrap(cls, object: APObject, actor_id: ObjectId) -> APCreate:
        """Wraps a given object into a Create activity.
        Only sets the "id" (to None), "actor", "object", and "type" fields.
        Does NOT set any of the "target", "to" etc. fields.

        Args:
            object (APObject): Object to be wrapped.
            actor_id (str): actor id that is performing this activity.

        Returns:
            APCreate: the create activity.
        """
        activity = APCreate(actor=actor_id, object=object)

        # copy recipients & visibility from the object to the activity
        for key in ["to", "cc", "bto", "bcc", "audience", "visibility"]:
            value = getattr(object, key, None)
            if value:
                setattr(activity, key, value)

        return activity

    @staticmethod
    async def store_object_locally(handler: ActivityHandler, obj: APObject, actor_id: ObjectId) -> APObject:
        # drop the id from the object (we create our own)
        obj.id = None

        # Attributed to creating user
        obj.attributed_to = actor_id

        # Clear the actor property
        if hasattr(obj, "actor"):
            setattr(obj, "actor", None)

        # For non-transient objects, the server MUST attach an id to both the wrapping Create and its wrapped Object.
        # https://www.w3.org/TR/activitypub/#client-to-server-interactions
        object_id = await handler.storage.object.create(data=obj)

        # add the id to the object so that it can be dereferenced when being stored in the activity
        obj.id = object_id

        # Set the blocks and shares collection ids
        obj.likes = collection_id_from_name(id=object_id, name="likes")
        _ = await handler.collections.create_collection(obj.likes, attributed_to=obj.id)
        obj.shares = collection_id_from_name(id=object_id, name="shares")
        _ = await handler.collections.create_collection(obj.shares, attributed_to=obj.id)
        _ = await handler.storage.object.upsert(id=object_id, data=obj)

        # if the instance supports objects lists for actors, add the object to the objects collection
        actor = await handler.resolver.resolve_actor(actor_id=actor_id)
        if actor.objects:
            await handler.collections.add_to_collection(
                collection_id=actor.objects, items=object_id, access=AccessType.from_visibility(obj.visibility)
            )

        # coll_shares = await self.collections.get_collection_page(collection_id=obj.shares)
        return obj

    @override
    async def validate(self, activity: APActivity) -> APActivity:
        """
        Enhanced Create validation.

        Validates:
        - Activity structure
        - Object type support
        - Content rules
        - Media attachments
        - Rate limits
        """
        # Validate activity.object
        if activity.type != ActivityType.CREATE.value:
            raise ValidationError(f"Create: APActivity is not APCreate (is type {activity.type})")

        obj_dict = activity.object
        if isinstance(obj_dict, APObject):
            obj_dict = obj_dict.serialize()
        if not isinstance(obj_dict, dict):
            self.logger.warning(f"CreateHandler: object in activity was not an APObject (instead: {type(obj_dict)}): activity {activity.serialize()}")
            raise ValidationError(f"Object in CREATE did not have proper information")
        obj_type = cast(str, obj_dict.get("type", ""))

        if not obj_type or obj_type not in self.SUPPORTED_TYPES:
            raise ValidationError(f"Unsupported object type: {obj_type}")

        # Validate object model
        model_class = self.SUPPORTED_TYPES[obj_type]
        obj = model_class.model_validate(obj_dict)

        # Check actor permissions
        actor_id = dereference(activity.actor, key="id")
        if not actor_id:
            raise ValidationError("Create: APCreate.actor.id is None")

        actor = await self.resolver.resolve_actor(actor_id=actor_id)
        if not actor:
            raise ValidationError("Actor not found")

        # Attribute the object to the actor of Create
        if obj.attributed_to and obj.attributed_to != actor_id:
            raise ValidationError(f"Object in Create is attributed to the wrong actor: attributed_to={obj.attributed_to}, actor_id={actor.id}")
        obj.attributed_to = actor_id

        # Validate file hash
        await CreateHandler.validate_content_hash(self, obj)

        # Validate content
        obj = await self._validate_content(obj)

        # Visibility of the object cannot be less secure than activity
        object_access = AccessType.from_visibility(obj.visibility)
        activity_access = AccessType.from_visibility(activity.visibility)
        maximim_security_visibility = AccessType.maximum_security(object_access, activity_access).value

        obj.visibility = maximim_security_visibility

        # Important for "store_in_collection" of server to create the object at this point, and not too late in the handle_outbox method
        activity_object = await CreateHandler.store_object_locally(handler=self, obj=obj, actor_id=actor_id)
        activity.object = activity_object

        # Set the visibility of the activity to be at most the one from the object
        activity.visibility = maximim_security_visibility

        return activity

    @override
    async def process_outbox(self, activity: APActivity) -> APActivity:
        """
        Enhanced Create processing.

        Handles:
        - Object storage
        - Side effects
        - Notifications
        - Federation
        """
        # TODO: copy all receipients from activity to object, then back
        # A mismatch between addressing of the Create activity and its object is likely to lead to confusion. As such, a server SHOULD copy any recipients of the Create activity to its object upon initial distribution, and likewise with copying recipients from the object to the wrapping Create activity.
        # https://www.w3.org/TR/activitypub/#create-activity-outbox

        # create object locally
        actor_id = dereference(activity, key="actor")
        if not actor_id:
            raise HandlerError("Create activity does not have actor")

        obj = await self.resolver.resolve_object(activity.object)
        if not obj.id:
            raise HandlerError("obj.id is None")

        # Process mentions
        # TODO: implement mentions
        # if mentions := self._extract_mentions(obj):
        #    self._handle_mentions(mentions, activity)

        # Process attachments
        # TODO: implement attachments
        if attachment := dereference(obj, "attachment"):
            await self._process_attachments(attachment, obj.id)

        # Handle notifications
        # TODO: implement notifications
        # self._send_notifications(activity)

        return activity

    @override
    async def process_inbox(self, activity: APActivity) -> None | str:
        self.logger.info(f"CreateHandler: Received CREATE in inbox (activity id: {activity.id})", extra={"activity_id": activity.id})

        # Validate content proof if required in settings.
        if self.storage.settings.media.require_file_proof_in_inbox:
            self.logger.debug(f"Validate file proof in inbox", extra={"activity_id": activity.id})
            obj = await self.resolver.resolve_object(activity.object)
            await CreateHandler.validate_proof(self, obj)

        return None

    async def _process_attachments(self, attachment: ObjectId, object_id: ObjectId):
        # TODO: actually process attachments
        attachment = attachment
        object_id = object_id

    async def _validate_content(self, object: APDocument | APPublicKey) -> APDocument | APPublicKey:
        # TODO: implement more content validation
        return object

    @staticmethod
    async def validate_content_hash(handler: ActivityHandler, object: APObject):
        """If required in settings, check if the object has a "hash" field attached to it (without checking its validity).

        Args:
            object (APObject): Object that should have a 'hash' field set.

        Raises:
            SecurityError: If required in settings and there is no hash field set to a string value.
        """
        if handler.storage.settings.media.require_content_hash:
            if not object.hash:
                raise SecurityError(f"Could not find a 'hash' field attached to the {object.type} with id '{object.id}'")

    @staticmethod
    async def validate_proof(handler: ActivityHandler, object: APObject):
        """Verify the object has a DataIntegrityProof attached in a 'proof' field and the object.hash is properly signed with the given proof.

        Args:
            object (APObject): Object with 'hash' and 'proof' fields.

        Raises:
            SecurityError: If the proof verification fails.
        """
        if handler.storage.settings.media.require_file_proof_in_inbox and isinstance(object, APDocument):
            proof = object.proof
            # TODO: refactoring: move this check to DataIntegrityProof class
            if isinstance(proof, DataIntegrityProof) and proof.verification_method:
                if proof.cryptosuite == "RSA":
                    public_key = await handler.key_manager.get_public_key(key_id=str(proof.verification_method))
                    if public_key and object.hash:
                        if proof.verify(object.hash, key=public_key):
                            handler.logger.debug(f"Successfully verified object hash for object '{object.id}'")
                            return
                else:
                    handler.logger.warning(f"Could not verify the proof with its cryptosuite set to '{proof.cryptosuite}' - not implemented")
                    pass

            raise SecurityError(f"Could not find a proper DataIntegrityProof attached to the {object.type} with id '{object.id}'")



"""
server.py
ActivityPubServer, calls handlers for activities in inbox and outbox.
"""

from abc import ABC
import asyncio
from collections.abc import AsyncIterator, Awaitable
from http import HTTPStatus
from logging import Logger
from typing import Callable

from pydantic import SecretStr
import temporalio
from temporalio.client import Client
from temporalio.contrib.pydantic import PydanticPayloadConverter
import temporalio.converter

from phederation.cache import BaseCache
from phederation.cache.dictcache import DictCache
from phederation.collections.collections import CollectionManager
from phederation.collections.collections_nodb import CollectionManagerNoDB
from phederation.federation.delivery import DeliveryResult
from phederation.federation.ratelimit import RateLimiter
from phederation.handlers.accept import AcceptHandler
from phederation.handlers.addremove import AddRemoveHandler
from phederation.handlers.announce import AnnounceHandler
from phederation.handlers.base import ActivityHandler
from phederation.handlers.block import BlockHandler
from phederation.handlers.create import CreateHandler
from phederation.handlers.delete import DeleteHandler
from phederation.handlers.follow import FollowHandler
from phederation.handlers.like import LikeHandler
from phederation.handlers.maintenance import MaintenanceHandler
from phederation.handlers.media import MediaHandler
from phederation.handlers.migrate import MigrationHandler
from phederation.handlers.reject import RejectHandler
from phederation.handlers.undo import UndoHandler
from phederation.handlers.update import UpdateHandler
from phederation.models import (
    APActivity,
    APObject,
    APOrderedCollectionPage,
    ActorType,
    ValidCollection,
    dereference,
)
from phederation.models.activities import APCreate, APUpdate
from phederation.models.actors import APAccount, APActor
from phederation.models.objects import dereference_or_raise
from phederation.models.proofs import DataIntegrityProof
from phederation.security.actors import ActorManager
from phederation.security.key_management import KeyManager
from phederation.security.worker_encryption import EncryptionCodec
from phederation.storage.base import StorageBackend, fill_database_url_from_env
from phederation.storage.nosql import NoSQLStorageBackend
from phederation.utils import ObjectId, PUBLIC_URLS
from phederation.utils.base import (
    AccessType,
    ActivityType,
    NodeInfo,
    ObjectType,
    UrlType,
    actor_id_from_username,
    collection_id_from_name,
)
from phederation.utils.exceptions import (
    DeliveryError,
    HandlerError,
    SecurityError,
    ValidationError,
    catch_exceptions,
)
from phederation.utils.logging import configure_logger
from phederation.utils.settings import PhedSettings, StorageProviderType

from .delivery import ActivityDelivery
from .discovery import InstanceDiscovery
from .instance import APInstance
from .resolver import ActivityPubResolver


class ActivityPubServerBase(ABC):
    storage: StorageBackend
    key_manager: KeyManager
    actor_manager: ActorManager
    discovery: InstanceDiscovery
    delivery: ActivityDelivery
    rate_limiter: RateLimiter
    resolver: ActivityPubResolver
    instance: APInstance
    collections: CollectionManager
    worker_client: Client | None
    settings: PhedSettings


class ActivityPubServer(ActivityPubServerBase):
    """
    Base ActivityPub server implementation.

    Features:
     - Activity processing
     - Queue management
     - Storage operations
     - Federation protocols
    """

    def __init__(
        self,
        settings: PhedSettings,
        delivery_workflow_from_integration: Callable[[Client | None, PhedSettings], Awaitable[DeliveryResult]] | None = None,
    ):
        self.settings: PhedSettings = settings
        self.logger: Logger = configure_logger(__name__, prefix=settings.federation.logging_prefix)

        self.netloc: str = settings.domain.netloc
        self.base_url: str = settings.domain.base_url
        self.delivery_workflow_from_integration: Callable[[Client | None, PhedSettings], Awaitable[DeliveryResult]] | None = (
            delivery_workflow_from_integration
        )
        self.fernet_key_id: ObjectId | None = None

    def create_cache(self) -> DictCache | None:
        if self.settings.federation.cache_max > 0 and self.settings.federation.cache_ttl > 0:
            cache = DictCache(
                ttl=self.settings.federation.cache_ttl,
                max_size=self.settings.federation.cache_max,
            )
        else:
            cache = None
        return cache

    async def create_components(self):
        """Create all components of the server, including storage, key management, and caches."""
        # Create components, and then initialize them through the server
        # StorageBackend.register_provider(StorageProviders.Sqlite, provider_class=NeoSqliteBackend)
        StorageBackend.register_provider(StorageProviderType.Mongodb, provider_class=NoSQLStorageBackend)

        self.storage: StorageBackend = await StorageBackend.create(settings=self.settings)

        self.collections: CollectionManager = CollectionManagerNoDB(storage=self.storage, page_size=self.settings.storage.page_size)

        self.cache: BaseCache | None = self.create_cache()

        self.rate_limiter: RateLimiter = RateLimiter(cache=self.cache or DictCache(), settings=self.settings)

        self.discovery: InstanceDiscovery = InstanceDiscovery(
            settings=self.settings,
            cache=self.cache,
            request_timeout=self.settings.federation.delivery_timeout,
        )

        self.instance: APInstance = APInstance(
            settings=self.settings,
            instance_name=self.settings.federation.instance_name,
            collections=self.collections,
            discovery=self.discovery,
            settings_path="settings",
            cache=self.cache,
        )

        self.resolver: ActivityPubResolver = ActivityPubResolver(
            settings=self.settings, cache=self.cache, storage=self.storage, instance=self.instance
        )

        self.key_manager: KeyManager = KeyManager(
            settings=self.settings,
            cache=self.cache,
            resolver=self.resolver,
            storage=self.storage,
        )

        self.actor_manager: ActorManager = ActorManager(
            settings=self.settings, storage=self.storage, key_manager=self.key_manager, collections=self.collections
        )

        # will be created in initialize()
        self.worker_client: Client | None = None

        self.delivery: ActivityDelivery = ActivityDelivery(
            key_manager=self.key_manager,
            resolver=self.resolver,
            discovery=self.discovery,
            instance=self.instance,
            settings=self.settings,
            rate_limiter=self.rate_limiter,
        )

    HANDLERS: dict[ActivityType, type[ActivityHandler]] = {
        ActivityType.CREATE: CreateHandler,
        ActivityType.FOLLOW: FollowHandler,
        ActivityType.LIKE: LikeHandler,
        ActivityType.DELETE: DeleteHandler,
        ActivityType.ANNOUNCE: AnnounceHandler,
        ActivityType.UPDATE: UpdateHandler,
        ActivityType.ADD: AddRemoveHandler,
        ActivityType.REMOVE: AddRemoveHandler,
        ActivityType.ACCEPT: AcceptHandler,
        ActivityType.REJECT: RejectHandler,
        ActivityType.UNDO: UndoHandler,
        ActivityType.BLOCK: BlockHandler,
        ActivityType.MAINTENANCE: MaintenanceHandler,
        ActivityType.MIGRATE: MigrationHandler,
    }

    async def initialize_handlers(self):
        # Initialize handlers
        self.media_handler: MediaHandler = MediaHandler(
            settings=self.settings,
            resolver=self.resolver,
            storage_backend=self.storage,
        )
        self.handlers: dict[
            ActivityType,
            ActivityHandler,
        ] = {
            key: ActivityPubServer.HANDLERS[key](
                actor_manager=self.actor_manager,
                key_manager=self.key_manager,
                storage=self.storage,
                collections=self.collections,
                resolver=self.resolver,
                delivery=self.delivery,
            )
            for key in ActivityPubServer.HANDLERS.keys()
        }

    async def connect_to_temporal(self, key_manager: KeyManager) -> Callable[[], Awaitable[DeliveryResult]] | None:
        """Try to connect to the temporalio backend (three times).
        If sucessful, initialize the temporal client. When enabled in settings.worker.encrypt_in_transit, use the `EncryptionCodec`
        class with an encryption key generated by the key_manager. The key is stored in the database, so that all gunicorn workers have access to it.
        """
        if self.settings.worker.enabled:
            instance_actor_id = self.instance.actor_id
            fernet_key: bytes | None = None
            if self.settings.worker.encrypt_in_transit:
                self.fernet_key_id, fernet_key = await key_manager.generate_fernet_key(actor_id=instance_actor_id)
            client_url = fill_database_url_from_env(url=SecretStr("temporalio://${TEMPORAL_HOST}:${TEMPORAL_PORT}"))
            for k_try in range(3):
                try:
                    if self.settings.worker.encrypt_in_transit:
                        if not fernet_key:
                            raise SecurityError("Encryption key for temporalio messages is None")
                        payload_codec = EncryptionCodec(fernet_key=fernet_key, logging_prefix=self.settings.federation.logging_prefix)
                        data_converter = temporalio.converter.DataConverter(
                            payload_converter_class=PydanticPayloadConverter, payload_codec=payload_codec
                        )
                    else:
                        data_converter = temporalio.converter.DataConverter(payload_converter_class=PydanticPayloadConverter)

                    self.worker_client = await Client.connect(
                        client_url.get_secret_value(),
                        data_converter=data_converter,
                    )
                    self.logger.debug(f"Connected to temporal client.")
                    break
                except Exception as e:
                    self.logger.debug(f"Could not connect to temporal client. Error '{e}', trying {k_try+1}/3 times...")
                    await asyncio.sleep(3)
            if not self.worker_client:
                self.logger.error(f"Could not connect to temporal client, tried 3 times.")

            async def delivery_workflow_with_fixed_worker():
                if self.delivery_workflow_from_integration and self.worker_client:
                    return await self.delivery_workflow_from_integration(self.worker_client, self.settings)
                else:
                    raise DeliveryError("Worker client delivery is not set up")

            return delivery_workflow_with_fixed_worker

    async def initialize(self) -> None:
        """Initialize server components."""
        # First, we create all necessary components of the server (uninitialized).
        await self.create_components()
        # Then, we initialize them. Note that the order of initialization matters here,
        # because e.g. storage must be initialized before key management and the connection to temporal.
        await self.initialize_handlers()

        await self.storage.initialize()
        if self.settings.storage.database.clear_on_initialize:
            await self.storage.clear()

        await self.collections.initialize()
        await self.key_manager.initialize()
        await self.discovery.initialize()
        await self.instance.initialize()
        delivery_workflow = await self.connect_to_temporal(key_manager=self.key_manager)
        await self.delivery.initialize(
            local_delivery_inbox=self.handle_inbox, local_delivery_outbox=self.handle_outbox, delivery_workflow=delivery_workflow
        )

    async def create_instance_actors(self):
        # create user collection if it does not exist
        collection_users = collection_id_from_name(id=self.base_url, name="users")
        users = await self.storage.collection.read(id=collection_users)
        if not users:
            _ = await self.collections.create_collection(collection_id=collection_users)

        # create instance actor
        inbox_url = collection_id_from_name(id=self.instance.actor_id, name=ValidCollection.Inbox.value)
        outbox_url = collection_id_from_name(id=self.instance.actor_id, name=ValidCollection.Outbox.value)
        instance_actor = APActor(
            name=self.settings.federation.instance_name,
            preferred_username=self.settings.federation.instance_name,
            id=self.instance.actor_id,
            type=ObjectType.SERVICE.value,
            discoverable=False,
            inbox=inbox_url,
            outbox=outbox_url,
        )
        _ = await self.actor_manager.create_actor(activity=APCreate(object=instance_actor, actor=self.instance.actor_id))

        # create testuser
        testuser_name = "testuser"
        testuser_id = actor_id_from_username(base_url=self.settings.domain.base_url, username=testuser_name)
        inbox_url = collection_id_from_name(id=testuser_id, name=ValidCollection.Inbox.value)
        outbox_url = collection_id_from_name(id=testuser_id, name=ValidCollection.Outbox.value)
        instance_actor = APActor(
            name=testuser_name,
            preferred_username=testuser_name,
            id=self.instance.actor_id,
            type=ActorType.PERSON.value,
            discoverable=False,
            inbox=inbox_url,
            outbox=outbox_url,
        )
        _ = await self.actor_manager.create_actor(activity=APCreate(object=instance_actor, actor=self.instance.actor_id))

        self.logger.info(f"Created instance actors: {testuser_name}, {self.settings.federation.instance_name}")

    async def _clean_inbox_activity(self, activity: APActivity) -> APActivity:
        """Cleans attributes of the given activity.
        This assumes that the activity is only for the given actors inbox.
        Hence, it will remove all to, cc, bto, bcc properties, and
        set the "target" property to the "to" actor.

        Args:
            activity (APActivity): Activity to clean.

        Returns:
            APActivity: Cleaned activity.
        """
        if activity.to and len(activity.to) > 0:
            activity.target = activity.to[0]
        activity.to = None
        activity.cc = None
        activity.bto = None
        activity.bcc = None
        activity.audience = None
        return activity

    async def resolve_multiple(self, ids_to_resolve: list[ObjectId], actor_for_authorization: ObjectId | None = None):
        """Resolve multiple objects on the local instance and store them in a new collection.

        Args:
            ids_to_resolve (list[ObjectId]): The ids to resolve. Must be on the local instance, otherwise the object is not resolved.
            actor_for_authorization (ObjectId | None, optional): Actor for authorization of the individual resolve calls. Defaults to None.
        """
        # TODO: use a temporal worker to offload the resolving tasks?
        resolved_objects = await asyncio.gather(
            *[
                self.resolver.try_resolve_from_storage(url=object_id, url_type=UrlType.from_local_url(object_id))
                for object_id in ids_to_resolve
                if self.resolver.is_local(object_id)
            ]
        )

        # check access
        available_objects: list[APObject | DataIntegrityProof | NodeInfo] = []
        for obj in resolved_objects:
            if isinstance(obj, APObject):
                obj_resolved = await self.resolver.resolve_access_type(object=obj, actor_requesting_access=actor_for_authorization)
            else:
                obj_resolved = obj
            if obj_resolved and not isinstance(obj_resolved, AsyncIterator):
                available_objects.append(obj_resolved)

        return available_objects

    async def handle_pagination(
        self,
        collection_id: ObjectId,
        actor_id: ObjectId | None = None,
        actor_for_authorization: ObjectId | None = None,
        page: int | None = None,
    ) -> APOrderedCollectionPage:
        """
        Handles the pagination of the given collection of the given user.
        If the page is None, the first page of the collection is returned.
        Only one of actor_id or object_id can be set (the other must be None).
        If actor_id and object_id are None, this method can return "all users" or "all activities", based on the "collection" string.

        Args:
            collection_id (ObjectId): the id of the collection (followers, following, ...) to handle.
            actor_id (ObjectId): id of the actor which the collection belongs to (for objects: the attributed_to actor). Default: None
            actor_for_authorization (ObjectId | None): id of the actor to use for authorization. Default: None (only public items)
            page (int): The page number to load. Each page has a maximum of self.storage.settings.page_size items. Default: None (load first page).

        Returns:
            APOrderedCollection
        """

        access: AccessType = await self.determine_access(actor_id=actor_id, actor_for_authorization=actor_for_authorization)

        collection_obj = await self.collections.get_collection_page(
            collection_id=collection_id, attributed_to=actor_id, page=page or 1, access=access
        )

        return collection_obj

    async def determine_access(self, actor_id: ObjectId | None, actor_for_authorization: ObjectId | None) -> AccessType:
        """Determines the access level of objects, given an actor with the items, and an actor that tries for authorization.

        Args:
            actor_id (ObjectId | None): The actor somebody is trying to access items from.
            actor_for_authorization (ObjectId | None): The actor that is trying to access.

        Returns:
            AccessType: The access type (PUBLIC, PRIVATE, FOLLOWER).
        """
        access = AccessType.PUBLIC
        if actor_id and actor_id == actor_for_authorization:
            access = AccessType.PRIVATE
        if actor_for_authorization and actor_id and actor_id != actor_for_authorization:
            self.logger.warning(f"SERVER: trying follower collection, as {actor_id} != {actor_for_authorization}")
            actor = await self.resolver.resolve_actor(actor_id=actor_id)
            follower_collection_id = dereference(actor, ValidCollection.Followers.value)
            if follower_collection_id:
                followers = await self.resolver.resolve_items(collection_id=follower_collection_id)
                if actor_for_authorization in followers:
                    access = AccessType.FOLLOWER
        return access

    async def update_account(self, activity: APUpdate):
        """Update the account given in the .object field of the activity.

        Args:
            activity (APUpdate): An update activity with a user account to update.

        Returns:
            ObjectId: Id of the account that was updated.
        """
        # TODO: implement this
        activity = activity
        account = activity.object
        if not isinstance(account, APAccount) or account.id is None:
            raise ValidationError("User account does not have a proper id.")
        return account.id

    @catch_exceptions(HandlerError, "Failed handling activity in inbox")
    async def handle_inbox(self, activity: APActivity, inbox_owner_id: ObjectId | None = None) -> DeliveryResult:
        """
        Handle incoming activity.

        Args:
            activity: ActivityPub activity
            inbox_owner_id: ObjectId | None, optional id of the owner of the inbox.

        Returns:
            DeliveryResult: result of delivery to the inbox.

        Raises:
            HandlerError: If activity handling fails
        """
        self.logger.info(f"New activity {activity.type} to handle in server inbox of actor {inbox_owner_id}")
        if not inbox_owner_id:
            self.logger.debug(f"This is handling the SHARED inbox of the instance. Calling 'handle' for all actors in the receipient fields...")
            results: list[DeliveryResult] = []
            recipients: list[ObjectId] = await activity.get_recipients()

            # The server MUST remove the bto and/or bcc properties, if they exist, from the ActivityStreams object before delivery, but MUST utilize the addressing originally stored on the bto / bcc properties for determining recipients in delivery.
            # https://www.w3.org/TR/activitypub/#client-to-server-interactions
            activity = await ActivityDelivery.remove_hidden_recipients(activity)

            activity = await self._clean_inbox_activity(activity=activity)
            for receipient in recipients:
                activity.target = receipient
                if activity.target in PUBLIC_URLS:
                    receipient = self.instance.actor
                results.append(await self.handle_inbox(activity=activity, inbox_owner_id=receipient))
            return DeliveryResult.create_from_list(results)

        result = DeliveryResult()

        # attempts to deliver to an inbox on a non-federated server SHOULD result in a 405 Method Not Allowed response.
        # TODO: implement this behavior, including federated/non-federated servers.

        # verify properties
        if not activity.id:
            raise HandlerError(
                f"ActivityPubServer: Activity {activity.type} does not have an id.",
            )
        activity_id = activity.id

        sender_id = dereference(activity, "actor")
        if not sender_id:
            raise HandlerError(
                "There was not 'actor' in the activity in the inbox. Failing.",
            )

        # check if actor is banned on this instance, discard the message if they are
        if await self.instance.is_banned_actor(actor_id=sender_id):
            self.logger.debug(f"Received activity from banned actor '{inbox_owner_id}'. Discarding it.")
            result.status_code = HTTPStatus.UNAUTHORIZED
            result.error_message = "Unauthorized"
            result.failed = [activity_id]
            return result

        # Get activity type
        activity_type = ActivityType(activity.type)
        if not activity_type:
            raise HandlerError("ActivityPubServer: Activity has no type")

        # Check if the actor_id is actually an actor on this instance. If not, fail silently.
        actor = await self.storage.actor.read(id=inbox_owner_id)
        if not actor or not actor.local:
            self.logger.debug(f"ActivityPubServer: actor {inbox_owner_id} is not an actor on this instance. Failing silently.")
            result.failed.append(inbox_owner_id)
            result.error_message = f"Actor {inbox_owner_id} is not an actor on this instance."
            result.status_code = HTTPStatus.NOT_FOUND
            return result

        # store the activity locally so that we record what is sent to this instance.
        # It may already exist from other actors on this instance that also received it.
        # In that case, just ignore the "create".
        activity_id = await self.storage.activity.create(data=activity, raise_if_exists=False)

        # check if the receiving actor is blocking the sending actor. If yes, do not proceed.
        collection_blocks = collection_id_from_name(id=inbox_owner_id, name="blocks")
        if await self.collections.contains(collection_id=collection_blocks, item=sender_id):
            self.logger.debug(f"actor_id={inbox_owner_id} is blocking object_id={sender_id}, not proceeding with activity {activity.id}")

            result.failed = [activity_id]
            result.error_message = "Blocked"
            result.status_code = HTTPStatus.FORBIDDEN
            return result

        # This Activity is added by the receiver as an item in the inbox OrderedCollection.
        # https://www.w3.org/TR/activitypub/#server-to-server-interactions
        # Add to inbox collection of inbox_owner_id
        collection_inbox = collection_id_from_name(id=inbox_owner_id, name="inbox")
        access = AccessType.from_visibility(activity.visibility)
        _ = await self.collections.add_to_collection(collection_id=collection_inbox, items=activity.id, access=access)

        # Get appropriate handler
        handler = self.handlers.get(activity_type)
        if not handler:
            raise HandlerError(
                f"No handler for activity type {activity_type}",
            )

        # clean activity
        activity = await self._clean_inbox_activity(activity)

        # validate activity
        activity = await handler.validate(activity=activity)

        # log the activity as incoming
        await self.instance.log_incoming(activity=activity)

        # handle activity
        _ = await handler.process_inbox(activity=activity)

        self.logger.info(f"Handled inbox activity {activity_type}: {activity_id}")
        result.success = [activity_id]
        result.status_code = HTTPStatus.OK
        return result

    async def _proof_content_hash(self, object: APObject | None):
        if object is None:
            return None
        if self.key_manager.settings.media.proof_object_hash:
            if not object.hash:
                raise SecurityError(f"Could not find a 'hash' field attached to the {object.type} with id '{object.id}'")
            actor = await self.resolver.resolve_actor(dereference(object, "attributed_to"))
            actor_key = await self.key_manager.resolve_current_key(actor.public_key_as_ids())
            if not actor_key:
                raise SecurityError(f"Could not find a current public_key for the actor {actor.id} of object {object.type} with id '{object.id}'")
            object.proof = await self.key_manager.proof_string(string_to_sign=object.hash, private_key_id=actor_key.id)
            return object
        else:
            return object

    async def _wrap_in_create(self, object: APObject) -> APActivity:
        # Attributed to creating user
        object.attributed_to = dereference(data=object, key="actor")

        if not object.attributed_to:
            raise HandlerError(f"Object {object.id} does not have property attributed_to; cannot wrap in Create activity.")

        # Wrap the object in a Create activity. This copies all receipients from the object to the activity.
        activity_create = CreateHandler.wrap(object=object, actor_id=object.attributed_to)
        return activity_create

    @catch_exceptions(HandlerError, "Failed handling activity in outbox")
    async def handle_outbox(self, activity: APActivity | APObject) -> DeliveryResult:
        """
        Handle outgoing activity.

        Args:
            activity: ActivityPub activity

        Returns:
            Activity ID

        Raises:
            DeliveryError: If delivery fails
        """
        self.logger.info(f"New {activity.visibility.lower()} activity to handle in outbox: {activity.type}")

        actor_id = dereference(data=activity, key="actor")
        if not actor_id:
            raise HandlerError(f"Id of actor in outbox for activity {activity.type} is not set")

        # check if actor is banned on this instance, discard the message if they are
        if await self.instance.is_banned_actor(actor_id=actor_id):
            self.logger.debug(f"Received activity from banned actor '{actor_id}'. Discarding it.")
            result = DeliveryResult(failed=[actor_id], error_message="Unauthorized", status_code=HTTPStatus.UNAUTHORIZED)
            return result

        # rate limit actor outbox to avoid spammers
        if not self.rate_limiter.check_rate_limit(domain=actor_id, strip_domain=False):
            self.logger.debug(f"Actor '{actor_id}' is rate limited. Discarding activity {activity.type}.")
            result = DeliveryResult(failed=[actor_id], error_message="Rate limited", status_code=HTTPStatus.TOO_MANY_REQUESTS)
            return result

        # If not an activity: a single non-Activity object which will be wrapped in a Create activity by the server.
        # https://www.w3.org/TR/activitypub/#client-to-server-interactions
        if not activity.type or ObjectType.__contains__(activity.type):
            activity = await self._wrap_in_create(object=activity)

        if not isinstance(activity, APActivity):
            raise DeliveryError("Activity is not a proper APActivity")

        # All activities must have an actor doing them.
        if not activity.actor:
            raise DeliveryError(
                "ActivityPubServer: Activity did not have an 'actor' attribute set.",
            )

        if not ActivityType.__contains__(activity.type):
            raise DeliveryError(
                f"ActivityPubServer: Cannot handle activity with type {activity.type}",
            )

        if ActivityType(activity.type) == ActivityType.ANNOUNCE:
            to: list[ObjectId] = activity.to or []
            actor_obj = await self.resolver.resolve_actor(dereference(activity, key="actor"))
            if actor_obj.id != self.instance.actor_id:
                # only add the followers of actors that are not the instance actor
                if not actor_obj.followers:
                    self.logger.error(f"Actor {dereference(activity, key="actor")} in Announce activity does not have followers; dict={actor_obj}")
                    raise DeliveryError(
                        f"ActivityPubServer: Actor in Announce activity does not have followers",
                    )
                to.append(actor_obj.followers)
                activity.to = to

        recipient_ids = await activity.get_recipients()

        # The server MUST remove the bto and/or bcc properties, if they exist, from the ActivityStreams object before delivery, but MUST utilize the addressing originally stored on the bto / bcc properties for determining recipients in delivery.
        # https://www.w3.org/TR/activitypub/#client-to-server-interactions
        activity = await ActivityDelivery.remove_hidden_recipients(activity)

        # Servers MUST also exclude actors from the list which are the same as the actor of the Activity being notified about. That is, actors shouldn't have their own activities delivered to themselves.
        # https://www.w3.org/TR/activitypub/#delivery
        # actor_id = dereference(data=activity, key="actor")
        if actor_id in recipient_ids:
            recipient_ids.remove(actor_id)

        # Servers performing delivery to the inbox or shared_inbox properties of actors on other servers MUST provide the object property in the activity: Create, Update, Delete, Follow, Add, Remove, Like, Block, Undo. Additionally, servers performing server to server delivery of the following activities MUST also provide the target property: Add, Remove.
        # https://www.w3.org/TR/activitypub/#server-to-server-interactions
        # TODO: verify that the above is satisfied before delivery

        activity_type = ActivityType(activity.type)

        activity = await self.process_outbox_activity(activity)

        # log the activity as outgoing
        await self.instance.log_outgoing(activity=activity)

        # Deliver the activity to recipients (server2server), if any.
        # Note that this includes the bcc and bto ones, even though the activity itself does not have them set anymore.
        result = await self.delivery.deliver_activity(activity=activity, recipients=recipient_ids)

        self.logger.info(
            f"ActivityPubServer: Processed outbox activity {activity_type} with id {activity.id}, delivered to approx. {len(result.success)} recipients, including self. {len(result.failed)} deliveries failed."
        )
        return result

    async def process_outbox_activity(self, activity: APActivity, is_maintenance: bool = False):

        # If an Activity is submitted with a value in the id property, servers MUST ignore this and generate a new id for the Activity.
        # https://www.w3.org/TR/activitypub/#client-to-server-interactions
        activity.id = None

        activity_type = ActivityType(activity.type)
        actor_id = dereference_or_raise(data=activity, key="actor")

        if activity_type == ActivityType.MAINTENANCE and not is_maintenance:
            raise HandlerError(
                f"Cannot handle Maintenance activity: not in maintenance mode",
            )

        # validate the activity, dropping/changing attributes depending on the handler
        handler = self.handlers.get(activity_type, None)
        if not handler:
            raise HandlerError(
                f"ActivityPubServer: No handler for activity type: {activity_type}",
            )
        activity = await handler.validate(activity=activity)

        # Proof content if required by settings
        if activity_type == ActivityType.CREATE and self.settings.media.proof_object_hash:
            activity.object = await self._proof_content_hash(activity.get_object_or_None())

        # For non-transient objects, the server MUST attach an id to both the wrapping Create and its wrapped Object.
        # https://www.w3.org/TR/activitypub/#client-to-server-interactions
        activity.id = await self.storage.activity.create(data=activity)

        # Determine access level from visibility
        access: AccessType = AccessType.PUBLIC if activity.visibility.lower() == "public" else AccessType.PRIVATE

        # Store in the outbox collection
        collection_outbox = collection_id_from_name(id=actor_id, name="outbox")
        _ = await self.collections.add_to_collection(collection_id=collection_outbox, items=activity.id, access=access)

        # Handle the activity in the outbox of the local actor
        activity = await handler.process_outbox(activity=activity)

        return activity



from dataclasses import dataclass, field
from datetime import datetime, timezone
import enum
from typing import Annotated, Any, override
from urllib.parse import urlparse

import httpx
from pydantic import (
    AfterValidator,
    BaseModel,
    Field,
    HttpUrl,
    PlainSerializer,
    ValidationError,
)

from phederation.utils.serialization import ActivityPubBase


def serialize_datetime(datetime: datetime | None):
    if not datetime:
        return "null"
    return datetime.astimezone(tz=timezone.utc).isoformat()


APDateTime = Annotated[
    datetime,
    PlainSerializer(serialize_datetime, return_type=str),
]


class AccessType(enum.Enum):
    """Represents the type of access on and above which the object is accessible."""

    # everybody is allowed to access
    PUBLIC = "public"
    # Only follower of attributed_to actor are allowed to access
    # TODO: implement this
    FOLLOWER = "follower"
    # Only the attributed_to actor is allowed to access
    PRIVATE = "private"

    @override
    def __str__(self):
        return str(self.value)

    @staticmethod
    def to_number(_type: "AccessType"):
        if _type == AccessType.PUBLIC:
            return 0
        elif _type == AccessType.FOLLOWER:
            return 1
        else:
            return 2

    @staticmethod
    def minimum_security(t1: "AccessType", t2: "AccessType") -> "AccessType":
        """Computes the minimum access level between the two arguments.

        This makes the result THE LEAST secure.

        Examples:
            - minimum(PRIVATE, PUBLIC) == PUBLIC
            - minimum(FOLLOWER, PRIVATE) == FOLLOWER

        Args:
            t1 (AccessType): Access 1
            t2 (AccessType): Access 2

        Returns:
            AccessType: the least secure access of the two.
        """
        count1 = AccessType.to_number(t1)
        count2 = AccessType.to_number(t2)
        return t1 if count1 < count2 else t2

    @staticmethod
    def maximum_security(t1: "AccessType", t2: "AccessType") -> "AccessType":
        """Computes the maximum access level between the two arguments.

        This makes the result THE MOST secure.

        Examples:
            - minimum(PRIVATE, PUBLIC) == PRIVATE
            - minimum(FOLLOWER, PRIVATE) == PRIVATE

        Args:
            t1 (AccessType): Access 1
            t2 (AccessType): Access 2

        Returns:
            AccessType: the most secure access of the two.
        """
        count1 = AccessType.to_number(t1)
        count2 = AccessType.to_number(t2)
        return t1 if count1 > count2 else t2

    @staticmethod
    def from_visibility(visibility: str):
        for _type in [AccessType.PUBLIC, AccessType.FOLLOWER, AccessType.PRIVATE]:
            if visibility.lower() == _type.value.lower():
                return _type
        raise ValueError(f"Visibility '{visibility}' cannot be converted to AccessType")

    def is_accessible(self, by_access_type: "AccessType") -> bool:
        """Checks if the "self" access level is accessible by the provided access level.

        Example 1: "self=PRIVATE" and "by_access_type=FOLLOWER" will return False, because "PRIVATE" is more restrictive than "FOLLOWER".
        Example 2: "self=PUBLIC" and "by_access_type=FOLLOWER" will return True.

        Args:
            access_type (AccessType): Access level to check against.

        Returns:
            bool: True if the SELF access level is LESS OR AS restrictive than the by_access_type level.
        """
        if self == AccessType.PUBLIC:
            return True
        if self == AccessType.FOLLOWER and (by_access_type == AccessType.PRIVATE or by_access_type == AccessType.FOLLOWER):
            return True
        if self == AccessType.PRIVATE and by_access_type == AccessType.PRIVATE:
            return True
        return False


PUBLIC_URLS: list[str] = ["public", "Public", "as:Public", "https://www.w3.org/ns/activitystreams#Public"]


def validate_object_id(url: "None | ObjectId | HttpUrl | str | dict[str, Any] | ActivityPubBaseWithId") -> "ObjectId":
    if isinstance(url, ActivityPubBaseWithId):
        object_id = str(url.id)
    elif isinstance(url, dict):
        if "id" in url:
            object_id = url.get("id", None)
    if isinstance(url, str) or isinstance(url, HttpUrl):
        if url in PUBLIC_URLS:
            url = PUBLIC_URLS[-1]
        object_id = str(HttpUrl(url))
        return object_id
    raise ValueError(f"Could not resolve id in object '{url}'.")


"""
Represents an id of an APObject.
Basically this is just a string that is a resolveable, unique URI.
The main use case in this code base is to be able to resolve many different versions of this URI, e.g.
"just a string", "a dictionary with the id as a key", or even just "https://*object_id*/inbox".
"""
ObjectId = Annotated[str, AfterValidator(validate_object_id)]


class ActivityPubBaseWithId(ActivityPubBase):
    """Generic base class for ActivityPub classes, including an id of type ObjectId and a visibility field."""

    id: None | ObjectId = Field(default=None, description="Unique identifier")
    visibility: str = Field(default=AccessType.PUBLIC.value, description="Visibility of the object, useful for access management")


@dataclass
class RateLimit:
    """Rate limit settings for a single domain."""

    requests: int
    """ Limit of requests per period that are allowed from a given domain. """

    period: int
    """ Period (in seconds) for how long the number of requests are answered, until the number is reset. """


class InstanceEndpoints(BaseModel):
    shared_inbox: ObjectId | None = Field(default=None, description="Link to the shared inbox of the instance, if exposed")
    instance_actor_id: ObjectId | None = Field(default=None, description="Id of the instance actor, if exposed")
    resolve: ObjectId | None = Field(default=None, description="The /resolve endpoint for multiple objects, if exposed")


class NodeInfoUsageUsers(BaseModel):
    total: int = 0
    activeHalfyear: int = 0
    activeMonth: int = 0


class NodeInfoUsage(BaseModel):
    users: NodeInfoUsageUsers = field(default_factory=lambda: NodeInfoUsageUsers())
    localPosts: int = 0
    localComments: int = 0


class NodeInfo(ActivityPubBaseWithId):
    """NodeInfo data, containing statistics about the instance. Shared across the network."""

    type: str = "NodeInfo"
    software: dict[str, str] = field(default_factory=lambda: {})
    protocols: list[str] = field(default_factory=lambda: [])
    services: dict[str, list[str]] = field(default_factory=lambda: {})
    usage: NodeInfoUsage = field(default_factory=lambda: NodeInfoUsage())
    open_registrations: bool = False
    metadata: dict[str, Any] = field(default_factory=lambda: {})
    version: str = "2.1"  # the NodeInfo version that is supported
    updated_at: APDateTime = field(default_factory=lambda: datetime.now(timezone.utc))


class ObjectType(enum.Enum):
    """Object types."""

    OBJECT = "Object"

    NOTE = "Note"
    ARTICLE = "Article"
    DOCUMENT = "Document"
    PAGE = "Page"
    IMAGE = "Image"
    VIDEO = "Video"
    AUDIO = "Audio"
    TOMBSTONE = "Tombstone"

    ACTOR = "Actor"
    ACCOUNT = "Account"
    PROFILE = "Profile"
    USER = "User"

    PERSON = "Person"
    GROUP = "Group"
    ORGANIZATION = "Organization"
    SERVICE = "Service"
    APPLICATION = "Application"

    COLLECTION = "Collection"
    ORDERED_COLLECTION = "OrderedCollection"
    COLLECTION_PAGE = "CollectionPage"
    ORDERED_COLLECTION_PAGE = "OrderedCollectionPage"
    COLLECTION_ITEM = "CollectionItem"

    EVENT = "Event"
    PLACE = "Place"
    LINK = "Link"
    MENTION = "Mention"
    RELATIONSHIP = "Relationship"

    KEY = "Key"
    DATA_INTEGRITY_PROOF = "DataIntegrityProof"
    NODE_INFO = "NodeInfo"

    @override
    def __str__(self):
        return str(self.value)


class CollectionType(enum.Enum):
    """Collection types."""

    COLLECTION = "Collection"
    ORDERED_COLLECTION = "OrderedCollection"
    COLLECTION_PAGE = "CollectionPage"
    ORDERED_COLLECTION_PAGE = "OrderedCollectionPage"
    COLLECTION_ITEM = "CollectionItem"

    @override
    def __str__(self):
        return str(self.value)


class ActivityType(enum.Enum):
    """Activity types."""

    CREATE = "Create"
    FOLLOW = "Follow"
    LIKE = "Like"
    ANNOUNCE = "Announce"
    DELETE = "Delete"
    UPDATE = "Update"
    UNDO = "Undo"
    ACCEPT = "Accept"
    REJECT = "Reject"
    ADD = "Add"
    REMOVE = "Remove"
    BLOCK = "Block"

    MIGRATE = "Migrate"
    MAINTENANCE = "Maintenance"
    ACTIVITY_IN_QUEUE = "ActivityInQueue"

    @override
    def __str__(self):
        return str(self.value)


class UrlType(str, enum.Enum):
    Users = "users"
    Actors = "actors"
    Accounts = "accounts"
    Activities = "activities"
    ActivitiesIncoming = "activities_incoming"
    ActivitiesOutgoing = "activities_outgoing"
    Objects = "objects"
    Blocks = "blocks"
    Follows = "follows"
    Keys = "keys"
    Media = "media"
    Collections = "collections"
    CollectionItems = "collection_items"
    MessageQueue = "queue"
    Proofs = "proofs"
    NodeInfo = "nodeinfo"
    Maintenance = "maintenance"
    Migration = "migration"
    Instance = "instance"
    Resolve = "resolve"
    Unknown = ""

    @staticmethod
    def from_object_type(obj_type: str) -> "UrlType":
        if obj_type == "Key":
            return UrlType.Keys
        if obj_type == "Actor" or obj_type == "User":
            return UrlType.Users
        if obj_type == "DataIntegrityProof":
            return UrlType.Proofs
        if ActivityType.__contains__(obj_type):
            return UrlType.Activities
        if ObjectType.__contains__(obj_type):
            return UrlType.Objects
        return UrlType.Unknown
    
    @staticmethod
    def from_local_url(url: str) -> "UrlType":
        path = urlparse(url).path.split("/")
        path = [p for p in path if len(p) > 0]
        try:
            url_type = UrlType(path[0])
        except Exception:
            return UrlType.Unknown
        return url_type


def urljoin(*args: str) -> str:
    """
    Joins given arguments into an url. Trailing but not leading slashes are stripped for each argument.
    """
    args_clean = [a for a in args if len(str(a).strip("/")) > 0]

    return "/".join(map(lambda x: str(x).strip("/"), args_clean))


def assemble_id_url(type: UrlType, base_url: ObjectId, primary: str | None = None) -> ObjectId:
    if not primary:
        return urljoin(base_url, type.value)
    else:
        return urljoin(base_url, type.value, primary)


def actor_id_from_username(base_url: ObjectId, username: str) -> ObjectId:
    """
    Standardized way to get an actor id from a domain and username.
    """
    return assemble_id_url(type=UrlType.Users, base_url=base_url, primary=username)


def username_from_actor_id(actor_id: ObjectId):
    path_split = urlparse(actor_id).path.split("/")
    if len(path_split) < 3:
        raise ValidationError(f"Actor_id {actor_id} cannot be split properly to obtain username")
    username = path_split[2]
    return username


def collection_id_from_name(id: ObjectId, name: str, page: int | None = None) -> ObjectId:
    """Creates a resolvable id for a collection with given name."""
    if not page:
        return urljoin(id, name)
    else:
        return f"{urljoin(id, name)}?page={page}"


async def async_request(url: str, headers: dict[str, str], timeout: int):
    async with httpx.AsyncClient() as client:
        return await client.get(str(url), headers=headers, timeout=timeout)


@dataclass
class ResolverResult:
    """Result of resolving a given object or url."""

    content: str | dict[str, str] | None = None
    """Message returned by the service, if any"""

    status_code: None | int = None
    """Status code returned by the resolved Url, or None if there was a non-HttpStatus related error."""

    error_message: None | str = None
    """Error message for a failed resolve."""

    def get(self, key: str):
        if self.content and isinstance(self.content, dict) and key in self.content:
            return self.content[key]



from http import HTTPStatus
from typing import Annotated, override

from fastapi import Depends, HTTPException, Path, Query, Security
from fastapi.encoders import jsonable_encoder
from fastapi.responses import JSONResponse
from pydantic import ValidationError

from phederation.api.routes.base import BaseRoute
from phederation.models import (
    APActivity,
    APActor,
    APOrderedCollection,
    APOrderedCollectionPage,
    APUpdate,
    APUser,
    ErrorMessageActivityHandler,
    ErrorMessageActivityValidation,
    ErrorMessageActorUnauthorized,
    RelativeLink,
    RelativeLinkTemplate,
    ValidCollection,
)
from phederation.security.authentication import UserInfo
from phederation.utils import ObjectId
from phederation.utils.base import actor_id_from_username, collection_id_from_name
from phederation.utils.exceptions import (
    ActivityPubException,
    AuthenticationError,
    CollectionError,
    DeliveryError,
    HandlerError,
    UserError,
)


class ActorRoute(BaseRoute):

    async def _get_users(self, page: int | None = None) -> APOrderedCollectionPage:
        """Get all publicly available users.

        Note that access level to the entire users list does not use authorization,
        which means it is not possible to count the private users.

        Args:
            page (int | None, optional): The page to access. Defaults to None (first page).

        Returns:
            APOrderedCollectionPage: A paginated collection of all public users on the instance.
        """
        collection_id = collection_id_from_name(id=self.server.base_url, name="users")
        collection = await self.server.handle_pagination(collection_id=collection_id, page=page)
        return collection

    async def _get_user(
        self, actor: APActor, actor_for_authorization: ObjectId | None = None  # pyright: ignore[reportUnusedParameter]
    ) -> APUser | None:
        """Return the APActor information for the given username.

        The actor_for_authorization is unused, because even actors with discoverable=False
        must be accessible by their username / actor_id. Otherwise, http signature authentication
        (and more) would not be possible for these "private actors", because their public keys are not accessible.

        Args:
            username (str): username of the actor to return.
            actor_for_authorization (_type_, optional): actor.id of an actor trying to get access. Defaults to None, Unused.

        Returns:
            APUser | None: The APActor object for the given username, if it exists on the instance.
        """
        if actor.id is None:
            return None

        public_key = actor.public_key_as_ids()
        if not public_key:
            self.logger.error(f"Actor for id {actor.id} does not have a public_key.")

        user_dict = actor.serialize()
        user_dict["public_key"] = public_key
        user_dict["links"] = [
            RelativeLink(rel="self", type="application/activity+json", href=actor.id).model_dump(),
            RelativeLinkTemplate(
                rel="http://ostatus.org/schema/1.0/subscribe",
                template=f"{self.api.settings.domain.base_url}/authorize_interaction?uri={{uri}}",
            ).model_dump(),
        ]
        user_dict["preferred_username"] = actor.preferred_username
        try:
            user = APUser.deserialize(user_dict)
        except ValidationError as e:
            self.logger.warning(f"Could not deserialize APUser from actor '{actor.id}' in ActorRoute, error {e}. Data={user_dict}")
            return None

        if not isinstance(user, APUser):
            self.logger.warning(f"Could not deserialize APUser from actor '{actor.id}' in ActorRoute")
            return None

        return user

    async def _get_user_collection(
        self, actor_id: ObjectId | None, collection: str, page: int | None = None, actor_for_authorization: ObjectId | None = None
    ) -> APOrderedCollection:
        if not actor_id:
            raise CollectionError("Actor id is none")
        collection_id = collection_id_from_name(id=actor_id, name=collection)
        collection_obj = await self.server.handle_pagination(
            collection_id=collection_id,
            actor_id=actor_id,
            page=page,
            actor_for_authorization=actor_for_authorization,
        )
        return collection_obj

    @override
    def setup(self):

        @self.router.get(
            "/users",
            tags=["users"],
            response_model_exclude_none=True,
        )
        async def get_user_list(  # pyright: ignore[reportUnusedFunction]
            page: Annotated[int | None, Query(..., description="The page of the collection to load (if any)")] = None,
        ) -> APOrderedCollectionPage:
            """Get paginated OrderedCollection of all local and public users."""
            return await self._get_users(page=page)

        @self.router.put(
            "/users",
            tags=["users"],
            response_model_exclude_none=True,
        )
        async def create_new_user(  # pyright: ignore[reportUnusedFunction]
            user_info: Annotated[UserInfo, Security(self.middleware.authorization, scopes=["user"])],
        ):
            """Create a new user.
            The authorization route to authorize the user will already create them on the instance,
            so this route is only to set specific attributes that do not come from keycloak.
            """
            assert user_info.actor_id, "user_info actor id is not set"

            self.logger.info(f"Received new user account info to register user '{user_info.full_name}'")

            try:
                # update the account information
                account = await self.server.storage.account.read(user_info.actor_id)
                if not account:
                    raise UserError(f"Could not find account for user {user_info.actor_id}")
                account.email = user_info.email

                update_account = APUpdate(
                    object=account,
                    actor=user_info.actor_id,
                )
                _ = await self.server.update_account(activity=update_account)
            except ActivityPubException as e:
                self.logger.error(f"ActorRoute handler error in create_new_user: {e.message}")
                raise e

            return JSONResponse(
                content=jsonable_encoder({"Status": "User created successfully"}),
                status_code=HTTPStatus.OK,
                media_type="application/activity+json",
            )

        @self.router.delete("/users/{username}", tags=["users"])
        async def delete_user(  # pyright: ignore[reportUnusedFunction]
            username: str, user_info: Annotated[UserInfo, Security(self.middleware.authorization, scopes=["user"])]
        ):
            """Delete the given user (if authenticated as them)."""
            if username != user_info.preferred_username:
                raise UserError("User is not authenticated with the username of the actor to delete")

            actor_id = actor_id_from_username(username=username, base_url=self.api.settings.domain.base_url)

            try:
                await self.server.actor_manager.delete_actor(actor_id)
                return JSONResponse(
                    content={"Status": "OK", "detail": "Actor deleted."},
                    status_code=HTTPStatus.OK,
                    media_type="application/activity+json",
                )
            except ActivityPubException as e:
                self.logger.error(f"ActorRoute: could not delete actor, exception {e.message}.")
                return JSONResponse(
                    content={"Status": "Error", "detail": "Actor could not be deleted. " + (e.user_message or "")},
                    status_code=HTTPStatus.BAD_REQUEST,
                    media_type="application/activity+json",
                )

        @self.router.get(
            "/users/{username}",
            tags=["users"],
            response_model=APUser,
            response_model_exclude_none=True,
        )
        async def get_user_route(  # pyright: ignore[reportUnusedFunction]
            actor: Annotated[APActor, Depends(self.middleware.authentication.verify_local_username)],
            user_info: Annotated[UserInfo | None, Security(self.middleware.authentication, scopes=["user"])],
        ) -> APUser:
            """Handle user data requests."""
            if user := await self._get_user(actor, actor_for_authorization=user_info.actor_id if user_info else None):
                return user
            raise HTTPException(status_code=HTTPStatus.GONE)

        @self.router.post(
            "/users/{username}/inbox",
            tags=["users"],
            responses={
                HTTPStatus.BAD_REQUEST: {
                    "model": ErrorMessageActivityValidation,
                    "description": ErrorMessageActivityValidation().description,
                },
                HTTPStatus.INTERNAL_SERVER_ERROR: {
                    "model": ErrorMessageActivityHandler,
                    "description": ErrorMessageActivityHandler().description,
                },
                HTTPStatus.UNAUTHORIZED: {
                    "model": ErrorMessageActorUnauthorized,
                    "description": ErrorMessageActorUnauthorized().description,
                },
            },
        )
        async def post_user_inbox(  # pyright: ignore[reportUnusedFunction]
            actor: Annotated[APActor, Depends(self.middleware.authentication.verify_local_username)],
            activity: Annotated[APActivity, Security(self.middleware.activity_from_http, scopes=["user"])],
        ):
            """Handle actor inbox requests."""
            result = "Result unset"
            try:
                result = await self.server.handle_inbox(activity=activity, inbox_owner_id=actor.id)
                if result.success and actor.id:
                    activity_id = str(result.success[0])
                else:
                    raise HandlerError(f"Could not handle inbox of actor {actor.id}")
            except ActivityPubException as e:
                self.logger.debug(f"Handle inbox failed with {type(e).__name__} {e.message}. Failing (HttpStatus.INTERNAL_SERVER_ERROR).")
                return JSONResponse(
                    content=e.user_message,
                    status_code=HTTPStatus.INTERNAL_SERVER_ERROR,
                    media_type="application/activity+json",
                )

            # as we have verified the remote url through HTTP signature, we can also tell the remote instance its current rate limit here
            response_headers = {"Location": activity_id}
            response_headers = self.middleware.rate_limiter.update_headers(domain=activity_id, headers=response_headers)
            self.logger.debug(f"Updated response headers for rate limit: {response_headers}")

            # Servers MUST return a 201 Created HTTP code, and unless the activity is transient, MUST include the new id in the Location header.
            # https://www.w3.org/TR/activitypub/#client-to-server-interactions
            return JSONResponse(
                content={"Status": "OK"},
                status_code=HTTPStatus.CREATED,
                media_type="application/activity+json",
                headers=response_headers,
            )

        @self.router.post("/users/{username}/outbox", tags=["users"])
        async def post_user_outbox(  # pyright: ignore[reportUnusedFunction]
            activity: APActivity,
            user_info: Annotated[UserInfo, Security(self.middleware.authorization, scopes=["user"])],
            actor: Annotated[APActor, Depends(self.middleware.authentication.verify_local_username)],
        ):
            """
            Handle user outbox requests (client-to-server).

            The POST is secured by authentication, the request is not verified.
            This is necessary because then clients do not need to sign the headers (they cannot, because they do not have the private keys),
            but clients need to authenticate using Bearer token (Oauth2).
            """
            if actor.id is None:
                raise AuthenticationError(f"Actor does not have an id.")

            if user_info.preferred_username != actor.preferred_username:
                raise AuthenticationError(f"User with username '{actor.preferred_username}' is not the authenticated user.")

            # set actor who this was posted to as the owner of the activity.
            activity.actor = actor.id

            try:
                result = await self.server.handle_outbox(activity=activity)
                success_id = result.get_success_id()
                if not success_id:
                    raise DeliveryError(f"Delivery to outbox failed: result={result}, actor.id={actor.id}, activity={activity.type}")
            except ActivityPubException as e:
                self.logger.debug(f"Handle outbox failed with {type(e).__name__} {e.message}. Failing silently (HttpStatus.CONTINUE).")
                e.status_code = HTTPStatus.CONTINUE
                return JSONResponse(
                    content={"Status": "Error", "description": e.user_message},
                    status_code=HTTPStatus.CONTINUE,
                    media_type="application/activity+json",
                )

            # Servers MUST return a 201 Created HTTP code, and unless the activity is transient, MUST include the new id in the Location header.
            # https://www.w3.org/TR/activitypub/#client-to-server-interactions
            return JSONResponse(
                content={"Status": "OK"},
                status_code=HTTPStatus.CREATED,
                media_type="application/activity+json",
                headers={"Location": success_id},
            )

        @self.router.get(
            "/users/{username}/blocks",
            response_model_exclude_none=True,
            tags=["users"],
            dependencies=[Security(self.middleware.authorization, scopes=["user"])],  # make blocks only available after authentication
        )
        async def get_user_blocks(  # pyright: ignore[reportUnusedFunction]
            actor: Annotated[APActor, Depends(self.middleware.authentication.verify_local_username)],
            user_info: Annotated[UserInfo | None, Security(self.middleware.authentication, scopes=["user"])],
            page: Annotated[int | None, Query(..., description="The page of the collection to load (if any)")] = None,
        ) -> APOrderedCollection:
            """Return list of blocked objects (actors) by the authenticated actor.
            This will return all objects in the given users "blocks" collection, if available.
            """
            return await self._get_user_collection(
                actor_id=actor.id,
                collection=ValidCollection.Blocks.value,
                page=page,
                actor_for_authorization=user_info.actor_id if user_info else None,
            )

        @self.router.get(f"/users/{{username}}/{ValidCollection.Collections.value}/{{collection}}", response_model_exclude_none=True, tags=["users"])
        async def get_user_defined_collection(  # pyright: ignore[reportUnusedFunction]
            actor: Annotated[APActor, Depends(self.middleware.authentication.verify_local_username)],
            collection: Annotated[str, Path(description="Name of collection created by actor")],
            user_info: Annotated[UserInfo | None, Security(self.middleware.authentication, scopes=["user"])],
            page: Annotated[int | None, Query(..., description="The page of the collection to load (if any)")] = None,
        ) -> APOrderedCollection | APOrderedCollectionPage:
            """Handle collection (GET) requests for user-defined collections.
            This will return public objects in the given users "collection", if available.
            If there are too many objects, it returns an OrderedCollectionPage of the collection.
            """
            # TODO: refactor so that "get_user_collection" is used instead (combined)
            collection_path = f"{ValidCollection.Collections.value}/{collection}"
            self.logger.debug(f"Accessing actor collection {collection_path}, page {page if page else "not set"}")
            collection_obj = await self._get_user_collection(
                actor_id=actor.id, collection=collection_path, page=page, actor_for_authorization=user_info.actor_id if user_info else None
            )
            return collection_obj

        @self.router.get("/users/{username}/{collection}", response_model_exclude_none=True, tags=["users"])
        async def get_user_collection(  # pyright: ignore[reportUnusedFunction]
            actor: Annotated[APActor, Depends(self.middleware.authentication.verify_local_username)],
            collection: Annotated[ValidCollection, Path(description="Collection name. Can be one of `ValidCollection`")],
            user_info: Annotated[UserInfo | None, Security(self.middleware.authentication, scopes=["user"])],
            page: Annotated[int | None, Query(..., description="The page of the collection to load (if any)")] = None,
        ) -> APOrderedCollection | APOrderedCollectionPage:
            """Handle certain collection (GET) requests.
            This will return public objects in the given users "collection", if available.
            If there are too many objects, it returns an OrderedCollectionPage of the collection.
            """
            self.logger.debug(f"Accessing actor collection '{collection.value}', page {page if page else "not set"}")
            collection_obj = await self._get_user_collection(
                actor_id=actor.id, collection=collection.value, page=page, actor_for_authorization=user_info.actor_id if user_info else None
            )
            return collection_obj

        @self.router.get(
            "/actor",
            tags=["instance"],
            response_model=APUser,
            response_model_exclude_none=True,
        )
        async def get_instance_actor() -> APUser:  # pyright: ignore[reportUnusedFunction]
            """Get instance actor."""
            actor = await self.middleware.authentication.verify_local_username(self.api.settings.federation.instance_name)
            if user := await self._get_user(actor):
                return user
            raise HTTPException(status_code=HTTPStatus.GONE)

        @self.router.get(
            "/inbox",
            tags=["instance"],
            response_model_exclude_none=True,
        )
        async def get_shared_inbox(page: int | None = None) -> APOrderedCollection:  # pyright: ignore[reportUnusedFunction]
            """Get activities from the shared inbox."""
            if "sharedInbox" not in self.api.settings.federation.endpoints:
                raise HTTPException(status_code=HTTPStatus.METHOD_NOT_ALLOWED, detail="Instance does not support the sharedInbox endpoint")

            actor = await self.middleware.authentication.verify_local_username(self.api.settings.federation.instance_name)
            inbox = await self._get_user_collection(
                page=page,
                actor_id=actor.id,
                collection="inbox",
                actor_for_authorization=actor.id,  # this makes the inbox as accessible as the instance actor
            )
            self.logger.debug(f"InstanceRoute: accessing shared inbox of instance {self.api.settings.federation.instance_name}.")
            return inbox



from datetime import datetime, timezone
import enum
import json
from logging import Logger
import os
from typing import Any, cast
from urllib.parse import urlparse

from pydantic import HttpUrl

from phederation.cache import BaseCache
from phederation.cache.base import WithCache, with_cache
from phederation.collections.collections import CollectionManager
from phederation.federation.discovery import InstanceDiscovery, InstanceInfo
from phederation.models import RelativeLink, dereference
from phederation.models.activities import APActivity
from phederation.models.actors import APActor
from phederation.utils import ObjectId
from phederation.utils.base import (
    AccessType,
    NodeInfo,
    UrlType,
    actor_id_from_username,
    assemble_id_url,
    collection_id_from_name,
    urljoin,
)
from phederation.utils.logging import configure_logger
from phederation.utils.settings import PhedSettings
from phederation.utils.version import PHEDERATION_SOFTWARE_NAME, PHEDERATION_VERSION


class InstanceCollectionType(str, enum.Enum):
    Federated = "federated"
    FederatedInstanceActors = "federated_instance_actors"
    Blocked = "blocked"
    BannedActors = "bannedactors"
    FederatedSharedInboxes = "federated_shared_inboxes"
    MaintenanceCommands = UrlType.Maintenance.value
    ActivitiesIncoming = UrlType.ActivitiesIncoming.value
    ActivitiesOutgoing = UrlType.ActivitiesOutgoing.value


class APInstance(WithCache):
    """
    Holds information about the current instance in the ActivityPub federation.
    """

    def __init__(
        self,
        settings: PhedSettings,
        instance_name: str,
        collections: CollectionManager,
        discovery: InstanceDiscovery,
        settings_path: str,
        cache: BaseCache | None = None,
    ):
        self.cache: BaseCache | None = cache
        self.settings: PhedSettings = settings
        self.settings_path: str = settings_path
        self.discovery: InstanceDiscovery = discovery
        self.collections: CollectionManager = collections
        self.instance_name: str = instance_name

        # this is the instance actor, a SERVICE actor
        self.actor_id: ObjectId = actor_id_from_username(base_url=self.settings.domain.base_url, username=instance_name)
        self.netloc: str = urlparse(self.actor_id).netloc

        # this is the instance id, where special collections can be obtained in admin mode
        self.instance_id: ObjectId = assemble_id_url(type=UrlType.Instance, base_url=self.settings.domain.admin_url)
        self.federated_instances: ObjectId = collection_id_from_name(id=self.instance_id, name=InstanceCollectionType.Federated.value)
        self.federated_instance_actors: ObjectId = collection_id_from_name(
            id=self.instance_id, name=InstanceCollectionType.FederatedInstanceActors.value
        )
        self.blocked_instances: ObjectId = collection_id_from_name(id=self.instance_id, name=InstanceCollectionType.Blocked.value)
        self.banned_actors: ObjectId = collection_id_from_name(id=self.instance_id, name=InstanceCollectionType.BannedActors.value)
        self.maintenance_commands: ObjectId = collection_id_from_name(id=self.instance_id, name=InstanceCollectionType.MaintenanceCommands.value)
        self.activities_incoming: ObjectId = collection_id_from_name(id=self.instance_id, name=InstanceCollectionType.ActivitiesIncoming.value)
        self.activities_outgoing: ObjectId = collection_id_from_name(id=self.instance_id, name=InstanceCollectionType.ActivitiesOutgoing.value)
        self._node_info: NodeInfo = NodeInfo()

        # TODO: enable to set this in settings
        self.federated_instances_access: AccessType = AccessType.PRIVATE
        self.blocked_instances_access: AccessType = AccessType.PRIVATE
        self.banned_actors_access: AccessType = AccessType.PRIVATE

        self.logger: Logger = configure_logger(name=__name__, prefix=self.settings.federation.logging_prefix)

    async def initialize(self):
        self._node_info = NodeInfo(
            software={
                "name": PHEDERATION_SOFTWARE_NAME,
                "version": PHEDERATION_VERSION,
                "homepage": self.settings.domain.url,
            },
            protocols=["activitypub"],
            services={"inbound": [], "outbound": []},
            open_registrations=False,
            metadata={"nodeName": self.instance_name, "nodeDescription": f"{self.instance_name} is a {PHEDERATION_SOFTWARE_NAME} software instance."},
            updated_at=datetime.now(timezone.utc),
        )

        self.federated_instances = await self.collections.create_collection(collection_id=self.federated_instances)
        self.federated_instance_actors = await self.collections.create_collection(collection_id=self.federated_instance_actors)
        self.blocked_instances = await self.collections.create_collection(collection_id=self.blocked_instances)
        self.banned_actors = await self.collections.create_collection(collection_id=self.banned_actors)
        self.maintenance_commands = await self.collections.create_collection(collection_id=self.maintenance_commands)
        self.activities_incoming = await self.collections.create_collection(collection_id=self.activities_incoming)
        self.activities_outgoing = await self.collections.create_collection(collection_id=self.activities_outgoing)

    async def log_incoming(self, activity: APActivity):
        if not activity.id:
            self.logger.debug(f"Could not log incoming {activity.type} activity: does not have an id")
        else:
            if self.settings.statistics.enabled and self.settings.statistics.log_activities_incoming:
                await self.collections.add_to_collection(
                    collection_id=self.activities_incoming, items=activity.id, access=AccessType.PRIVATE, add_only_once=False
                )

    async def log_outgoing(self, activity: APActivity):
        if not activity.id:
            self.logger.debug(f"Could not log outgoing {activity.type} activity: does not have an id")
        else:
            if self.settings.statistics.enabled and self.settings.statistics.log_activities_outgoing:
                self.logger.debug(f"Logging outgoing activity {activity.id}")
                await self.collections.add_to_collection(
                    collection_id=self.activities_outgoing, items=activity.id, access=AccessType.PRIVATE, add_only_once=False
                )

    async def federate(self, instance_id: ObjectId):
        """Given a generic url, saves the id in the list of federated instances, if it is not already in it.
        Also tries to discover the InstanceInfo using discover_instance, and (if accessible) stores the shared_inbox of the remote instance. This can then be used to deliver public objects to it.

        Args:
            instance_id (ObjectId): generic url, will be converted to scheme://netloc (no path, query, fragemnts) and processed.
        """
        parsed_url = urlparse(instance_id)
        instance_id = HttpUrl.build(scheme=parsed_url.scheme, host=parsed_url.netloc).__str__()
        # only federate other instances, not self
        if self.netloc != parsed_url.netloc:
            federated_instance_ids = await self._get_federated_instances(collection=self.federated_instances, access=self.federated_instances_access)
            if not instance_id in federated_instance_ids:
                self.logger.info(f"Federating new instance: {instance_id}, currently federated: {federated_instance_ids}")
                # Now discover information about the remote instance, and save its shared inbox, if available
                instance_info = await self.discovery.discover_instance(instance_id)
                instance_actor = dereference(instance_info.instance_actor, "id")
                if not instance_actor:
                    self.logger.warning(f"Instance actor of instance to federate ({instance_id}) is None")
                    return

                await self.collections.add_to_collection(
                    collection_id=self.federated_instances, items=instance_id, access=self.federated_instances_access, add_only_once=True
                )
                await self.collections.add_to_collection(
                    collection_id=self.federated_instance_actors, items=instance_actor, access=self.federated_instances_access, add_only_once=True
                )

                # clear cache to reload everything from storage next time it gets accessed
                if self.cache:
                    self.cache.clear(__name__)

    async def defederate(self, instance_id: ObjectId):
        parsed_url = urlparse(instance_id)
        instance_id = HttpUrl.build(scheme=parsed_url.scheme, host=parsed_url.netloc).__str__()
        # only defederate other instances, not self
        if self.netloc != parsed_url.netloc:
            if instance_id in await self._get_federated_instances(collection=self.federated_instances, access=self.federated_instances_access):
                await self.collections.remove_from_collection(
                    collection_id=self.federated_instances, items=instance_id, access=self.federated_instances_access
                )
                instance_info = await self.discovery.discover_instance(instance_id)
                instance_actor = dereference(instance_info.instance_actor, "id")
                if instance_actor:
                    await self.collections.remove_from_collection(
                        collection_id=self.federated_instance_actors,
                        items=instance_actor,
                        access=self.federated_instances_access,
                    )
                if self.cache:
                    self.cache.clear(__name__)

    @with_cache(__name__)
    async def _get_federated_instances(self, collection: ObjectId, access: AccessType = AccessType.PUBLIC) -> list[ObjectId]:
        federated = await self.collections.get_collection_items(collection_id=collection, access=access)
        return federated

    async def get_federated_instance_actors(self, access: AccessType = AccessType.PUBLIC) -> list[ObjectId]:
        """Get the instance actor of all federated instances.

        Args:
            access (AccessType, optional): The access type for the instances to retrieve. Defaults to AccessType.PUBLIC.

        Returns:
            list[ObjectId]: List of urls for instance actors all federated instances, depending on the access type set.
        """
        return await self._get_federated_instances(collection=self.federated_instance_actors, access=access)

    async def block_instance(self, url: ObjectId):
        """Adds the given instance url to the list of blocked instances.

        A blocked instance is not able to send or retrieve information from the current instance.

        Args:
            url (ObjectId): Url of the instance to block.
        """
        await self.defederate(url)
        await self.collections.add_to_collection(
            collection_id=self.blocked_instances, items=url, access=self.blocked_instances_access, add_only_once=True
        )
        if self.cache:
            self.cache.clear(__name__)

    async def unblock_instance(self, url: ObjectId):
        """Undo the block_instance command.

        Args:
            url (ObjectId): Url of the instance to unblock.
        """
        await self.collections.remove_from_collection(collection_id=self.blocked_instances, items=url, access=self.blocked_instances_access)
        if self.cache:
            self.cache.clear(__name__)

    async def _get_blocked_instances(self, access: AccessType = AccessType.PUBLIC) -> list[ObjectId]:
        return await self._get_federated_instances(collection=self.blocked_instances, access=access)

    async def is_blocked_instance(self, url: ObjectId):
        url_parsed = urlparse(url)
        url_netloc = url_parsed.netloc
        blocked_actors = await self._get_blocked_instances(access=self.blocked_instances_access)
        for blocked in blocked_actors:
            blocked_netloc = urlparse(blocked).netloc
            if blocked_netloc.strip("/") == url_netloc.strip("/"):
                return True
        # if the url is not blocked, federate it.
        instance_id = HttpUrl.build(scheme=url_parsed.scheme, host=url_parsed.netloc).__str__()
        await self.federate(instance_id=instance_id)
        return False

    async def ban_actor(self, actor_id: ObjectId):
        await self.collections.add_to_collection(
            collection_id=self.banned_actors, items=actor_id, access=self.banned_actors_access, add_only_once=True
        )
        if self.cache:
            self.cache.delete(self.banned_actors)

    async def unban_actor(self, actor_id: ObjectId):
        for _ in range(1000):  # remove all previous bans
            if await self.is_banned_actor(actor_id=actor_id):
                await self.collections.remove_from_collection(collection_id=self.banned_actors, items=actor_id, access=self.banned_actors_access)
            else:
                break
        if self.cache:
            self.cache.clear(__name__)

    async def is_banned_actor(self, actor_id: ObjectId | None):
        if not actor_id:
            return True
        return actor_id in await self._get_banned_actors(access=self.banned_actors_access)

    async def _get_banned_actors(self, access: AccessType = AccessType.PUBLIC) -> list[ObjectId]:
        return await self._get_federated_instances(collection=self.banned_actors, access=access)

    @property
    def actor(self):
        return self.actor_id

    @with_cache(prefix="nodeinfo getter")
    async def get_nodeinfo(self) -> NodeInfo:
        nodeinfos = await self.collections.storage.node_info.select(access=AccessType.PUBLIC)

        def sort_infos(key: NodeInfo):
            return key.updated_at

        nodeinfos.sort(key=sort_infos)
        if len(nodeinfos) > 0:
            self._node_info = nodeinfos[-1]
        return self._node_info

    def get_nodeinfo_links(self) -> dict[str, list[RelativeLink]]:
        nodeinfo_schema_version = "2.1"
        return {
            "links": [
                RelativeLink(
                    rel=urljoin("http://nodeinfo.diaspora.software/ns/schema/", nodeinfo_schema_version),
                    href=urljoin(self.settings.domain.base_url, "nodeinfo", nodeinfo_schema_version),
                )
            ]
        }

    def get_nodeinfo_schema(self) -> dict[str, Any]:
        with open(os.path.join(self.settings_path, "nodeinfo.schema.json"), "r+") as nodeinfo_schema_file:
            nodeinfo_schema = cast(dict[str, Any], json.load(nodeinfo_schema_file))
            nodeinfo_schema_id = cast(str, nodeinfo_schema.get("id", ""))
            nodeinfo_schema["id"] = nodeinfo_schema_id.replace("{DOMAIN}", self.settings.domain.url)
            return nodeinfo_schema

    async def get_instance_info(self) -> InstanceInfo:
        actor: APActor | None = await self.collections.storage.actor.read(id=self.actor_id)
        shared_inbox = urljoin(self.settings.domain.base_url, "inbox")
        features = {"activitypub": True, "webfinger": True, "nodeinfo": True, "shared_inbox": True, "media_proxy": True}
        endpoints = {
            "shared_inbox": shared_inbox,
            "nodeinfo": urljoin(self.settings.domain.base_url, "/.well-known/nodeinfo"),
            "webfinger": urljoin(self.settings.domain.base_url, "/.well-known/webfinger"),
        }
        if "resolve" in self.settings.federation.endpoints:
            endpoints["resolve"] = urljoin(self.settings.domain.base_url, UrlType.Resolve.value)
            features["resolve"] = True
        if "collections" in self.settings.federation.endpoints:
            features["collections"] = True
        return InstanceInfo(
            domain=self.settings.domain.base_url,
            nodeinfo=await self.get_nodeinfo(),
            endpoints=endpoints,
            features=features,
            instance_actor=actor.serialize() if actor else None,
            shared_inbox=shared_inbox,
            software_version=PHEDERATION_VERSION,
            last_updated=datetime.now(timezone.utc),
        )



import asyncio
from logging import Logger
import logging.config
import os
from typing import Any, cast

import yaml

from phederation.actions.worker import main_worker
from phederation.integration.base import BaseIntegration
from phederation.integration.integration import FastAPIIntegration
from phederation.utils.exceptions import SettingsError
from phederation.utils.logging import configure_logger
from phederation.utils.settings import PhedSettings


class Instance:
    integration: BaseIntegration | None = None
    main_instance_settings: PhedSettings | None = None

    def __init__(self, settings: PhedSettings):
        self.verify_settings(settings)
        self.settings: PhedSettings = settings.model_copy(deep=True)
        self.ssl_context: tuple[str, str] = (
            "PEM pub path not set",
            "PEM private path not set",
        )
        if self.settings.federation.admin_mode:
            self.configure_admin_settings()

        self.logger: Logger = configure_logger(__name__, prefix=self.settings.federation.logging_prefix)

        if self.settings.federation.admin_mode:
            self.logger.warning("Instance running in admin mode.")

    def verify_settings(self, settings: PhedSettings):
        if settings.domain.port == settings.domain.admin_port:
            raise SettingsError("The port of the main instance is set to be the same as the port in admin mode. This is now allowed")

    def configure_admin_settings(self):
        """Reconfigures the settings object to be used in admin mode.

        New configuration
         - Add an "ADMIN" tag ad the end of the logger prefix to log into a different file
         - Do not clear the database on start
         - Do not log any statistics / middleware
         - Do not use temporal workers for delivery

        The domain.url and domain.port are left the same as the main instance, to simulate
        that the admin instance is "the same" as the main one. This allows to easily resolve
        internal collections from the main instance.
        To be able to check the settings of the main instance, they are saved in "main_instance_settings"
        before changing them to the admin settings.
        """
        self.main_instance_settings = self.settings.model_copy(deep=True)
        self.settings.storage.database.clear_on_initialize = False
        self.settings.statistics.enabled = False
        self.settings.worker.enabled = False
        self.settings.federation.logging_prefix = f"{self.settings.federation.logging_prefix}-ADMIN"

    async def initialize(self):
        self.logger.debug(f"Initalizing instance.")

        self.integration = FastAPIIntegration(self.settings)
        await self.integration.initialize()

        self.logger.debug(f"Integration created.")

    async def run(self):
        await self.initialize()
        self.logger.debug(f"Instance initialized. Starting run.")
        if not self.integration:
            return None
        return self.integration.app

    def close(self):
        pass


async def main(env_file: str | None = None, settings_file: str | None = None, logging_file: str | None = None, admin_mode: str = "False"):

    if not env_file:
        env_file = ".env"
    if not settings_file:
        settings_file = "phederation.yaml"

    settings = PhedSettings(yaml_file=settings_file, env_file=env_file)

    if logging_file is None:
        with open(settings_file) as f:
            settings_config = cast(dict[str, Any], yaml.safe_load(f))
            logging_config = cast(dict[str, Any], settings_config["logging"])
    else:
        with open(logging_file) as f:
            logging_config = cast(dict[str, Any], yaml.safe_load(f))
    logging.config.dictConfig(config=logging_config)

    if admin_mode.lower() == "true":
        settings.federation.admin_mode = True

    instance = Instance(settings=settings)
    instance.logger.info(f"Using settings file '{os.path.basename(settings_file)}'")

    # for a WSGI application like gunicorn, we need to return the app, not run it
    await instance.initialize()
    if not instance.integration:
        return None
    return instance.integration.app


# Gunicorn entry point generator for phederation instance
def app(*args_app, **kwargs_app):  # pyright: ignore[reportUnknownParameterType, reportMissingParameterType, reportUnusedParameter]
    # Gunicorn CLI args are useless.
    # https://stackoverflow.com/questions/8495367/
    #
    # Start the application in modified environment.
    # https://stackoverflow.com/questions/18668947/
    #

    return asyncio.run(main(**kwargs_app))  # pyright: ignore[reportUnknownArgumentType]


# Gunicorn entry point generator for temporalio worker
def app_worker(*args_worker, **kwargs_worker):  # pyright: ignore[reportUnknownParameterType, reportMissingParameterType, reportUnusedParameter]

    return asyncio.run(main_worker(**kwargs_worker))  # pyright: ignore[reportUnknownArgumentType]


if __name__ == "__main__":
    _ = asyncio.run(main())



import base64
from datetime import datetime, timezone
import hashlib
from typing import Literal

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding, rsa
from cryptography.hazmat.primitives.asymmetric.ed25519 import (
    Ed25519PrivateKey,
    Ed25519PublicKey,
)
from pydantic import BaseModel, Field

from phederation.utils.base import APDateTime, ActivityPubBaseWithId, ObjectId
from phederation.utils.exceptions import SecurityError


class TokenRequest(BaseModel):
    username: str
    password: str


class TokenResponse(BaseModel):
    access_token: str
    token_type: str = "bearer"


def to_sha256(argument: str | bytes):
    """Computes the hexdigest of the argument using SHA256.

    Args:
        string_to_hash (str | bytes): String or bytes array to hash. The string will be encoded using utf-8.

    Returns:
        str: SHA256 of the encoded string.
    """
    if isinstance(argument, str):
        argument = argument.encode(encoding="utf-8")

    hash = hashlib.sha256()
    hash.update(argument)
    hash_sha256 = hash.hexdigest()
    return hash_sha256


def sign_rsa(string_to_sign: str, private_key: rsa.RSAPrivateKey):
    signature = private_key.sign(string_to_sign.encode(encoding="utf-8"), padding.PKCS1v15(), hashes.SHA256())
    return base64.b64encode(signature).decode(encoding="utf-8")


def verify_rsa(signature_string: str, string_to_verify: str, public_key: rsa.RSAPublicKey):
    signature = base64.b64decode(signature_string)
    try:
        _ = public_key.verify(
            signature=signature, data=string_to_verify.encode(encoding="utf-8"), padding=padding.PKCS1v15(), algorithm=hashes.SHA256()
        )
    except InvalidSignature:
        return False
    return True


def sign_ed25519(string_to_sign: str, private_key: Ed25519PrivateKey):
    signature = private_key.sign(string_to_sign.encode(encoding="utf-8"))
    return base64.b64encode(signature).decode(encoding="utf-8")


def verify_ed25519(signature_string: str, string_to_verify: str, public_key: Ed25519PublicKey):
    signature = base64.b64decode(signature_string)
    try:
        public_key.verify(signature, string_to_verify.encode(encoding="utf-8"))
    except InvalidSignature:
        return False
    return True


class DataIntegrityProof(ActivityPubBaseWithId):
    """A data integrity proof provides information about the proof mechanism,
    parameters required to verify that proof, and the proof value itself.
    When expressing a data integrity proof on an object, a "proof" property of type DataIntegrityProof MUST be used.

    https://w3c.github.io/vc-data-integrity/#dfn-data-integrity-proof
    """

    type: Literal["DataIntegrityProof"] = Field(
        default="DataIntegrityProof",
        description="Always set to DataIntegrityProof here, the cryptographySuite field specifies which type of proof.",
    )

    proof_purpose: Literal["assertionMethod"] | Literal["authentication"] = Field(
        default="assertionMethod",
        description="The reason the proof was created MUST be specified as a string that maps to a URL. The proof purpose acts as a safeguard to prevent the proof from being misused by being applied to a purpose other than the one that was intended. For example, without this value the creator of a proof could be tricked into using cryptographic material typically used to create a Verifiable Credential (assertionMethod) during a login process (authentication) which would then result in the creation of a verifiable credential they never meant to create instead of the intended action, which was to merely log in to a website.",
    )

    description: str | None = Field(
        default=None, description="Optional, human readable string to describe what the proof is doing or why it was issued."
    )

    verification_method: ObjectId | None = Field(
        default=None,
        description="A verification method is the means and information needed to verify the proof. If included, the value MUST be a string that maps to a [URL]. Inclusion of verificationMethod is OPTIONAL, but if it is not included, other properties such as cryptosuite might provide a mechanism by which to obtain the information necessary to verify the proof. Note that when verificationMethod is expressed in a data integrity proof, the value points to the actual location of the data; that is, the verificationMethod references, via a URL, the location of the public key that can be used to verify the proof. This public key data is stored in a controlled identifier document, which contains a full description of the verification method.",
    )

    cryptosuite: Literal["RSA"] | Literal["Ed25519Signature2020"] | None = Field(
        default=None,
        description="An identifier for the cryptographic suite that can be used to verify the proof. If the proof type is DataIntegrityProof, cryptosuite MUST be specified (RSA or Ed25519Signature2020); otherwise, cryptosuite MAY be specified. If specified, its value MUST be a string.",
    )

    created: APDateTime | None = Field(
        default=None,
        description="The date and time the proof was created is OPTIONAL.",
    )

    expires: APDateTime | None = Field(default=None, description="The expires property is OPTIONAL and, if present, specifies when the proof expires.")

    domain: str | None = Field(
        default=None,
        description="The domain property is OPTIONAL. It conveys one or more security domains in which the proof is meant to be used. If specified, the associated value MUST be either a string, or an unordered set of strings. A verifier SHOULD use the value to ensure that the proof was intended to be used in the security domain in which the verifier is operating. The specification of the domain parameter is useful in challenge-response protocols where the verifier is operating from within a security domain known to the creator of the proof.",
        examples=["domain.example", "https://domain.example:8443", "mycorp-intranet", "b31d37d4-dd59-47d3-9dd8-c973da43b63a"],
    )

    challenge: str | None = Field(
        default=None,
        description="A string value that SHOULD be included in a proof if a domain is specified. The value is used once for a particular domain and window of time. This value is used to mitigate replay attacks.",
        examples=["1235abcd6789", "79d34551-ae81-44ae-823b-6dadbab9ebd4", "ruby"],
    )

    proof_value: str | None = Field(
        default=None,
        description="A string value that expresses base-encoded binary data necessary to verify the digital proof using the verificationMethod specified. The value MUST use a header and encoding to express the binary data. The contents of this value are determined by a specific cryptosuite and set to the proof value generated by the 'Add Proof Algorithm' for that cryptosuite. Alternative properties with different encodings specified by the cryptosuite MAY be used, instead of this property, to encode the data necessary to verify the digital proof.",
    )

    previous_proof: str | list[str] | None = Field(
        default=None,
        description="The previousProof property is OPTIONAL. If present, it MUST be a string value or an unordered list of string values. Each value identifies another data integrity proof, all of which MUST also verify for the current proof to be considered verified.",
    )

    nonce: str | None = Field(
        default=None,
        description="An OPTIONAL string value supplied by the proof creator. One use of this field is to increase privacy by decreasing linkability that is the result of deterministically generated signatures.",
    )

    def sign(self, string_to_sign: str, key: rsa.RSAPrivateKey | Ed25519PrivateKey, public_key_id: ObjectId | None = None):
        if self.cryptosuite == "RSA" and isinstance(key, rsa.RSAPrivateKey):
            proofValue = sign_rsa(string_to_sign=string_to_sign, private_key=key)
        elif self.cryptosuite == "Ed25519Signature2020" and isinstance(key, Ed25519PrivateKey):
            proofValue = sign_ed25519(string_to_sign=string_to_sign, private_key=key)
        else:
            raise SecurityError(f"DataIntegrityProof signing: type {self.type} does not match the key type {type(key)}.")
        self.proof_value = proofValue
        self.created = datetime.now(tz=timezone.utc)
        if public_key_id:
            self.verification_method = public_key_id
        return self

    def verify(self, string_to_verify: str, key: rsa.RSAPublicKey | Ed25519PublicKey) -> bool:
        if self.proof_value and self.cryptosuite == "RSA" and isinstance(key, rsa.RSAPublicKey):
            return verify_rsa(signature_string=self.proof_value, string_to_verify=string_to_verify, public_key=key)
        elif self.proof_value and self.cryptosuite == "Ed25519Signature2020" and isinstance(key, Ed25519PublicKey):
            return verify_ed25519(signature_string=self.proof_value, string_to_verify=string_to_verify, public_key=key)
        raise SecurityError(f"DataIntegrityProof verification: type {self.type} does not match the key type {type(key)}.")



"""
tasks.py
All tasks related to federation:
- Updating node info (number of users, posts etc)
"""

from collections.abc import Sequence
from typing import Any, Callable, override
import uuid

from temporalio.worker import Worker
from temporalio.worker.workflow_sandbox import SandboxRestrictions, SandboxedWorkflowRunner

from phederation.federation.tasks import BackgroundTask


class WorkerTask(BackgroundTask):
    def __init__(self, workflows: Sequence[type], activities: Sequence[Callable[..., Any]]):
        super().__init__()
        self.workflows: Sequence[type] = workflows
        self.activities: Sequence[Callable[..., Any]] = activities

    @override
    async def on_startup(self):
        if not self.server:
            raise ValueError("Worker task not properly initialized")
        if not self.server.settings.worker.enabled or not self.server.worker_client:
            raise ValueError("Worker disabled")
        
        if not self.server.settings.worker.start_own_worker:
            self.logger.info("Starting an instance-owned worker is disabled, there must be workers started in separate processes.")
            return
        
        self.logger.debug("Worker starting...")
        worker = Worker(
            self.server.worker_client,
            identity=f"phederation-worker-{str(uuid.uuid4())[:8]}",
            workflow_runner=SandboxedWorkflowRunner(
                restrictions=SandboxRestrictions.default.with_passthrough_modules("pydantic", "annotated_types", "pydantic_core")
            ),
            task_queue=self.server.settings.worker.tasks_queue_name,
            workflows=self.workflows,
            activities=self.activities,
        )
        self.logger.debug("Worker started.")
        await worker.run()
        self.logger.debug("Worker shutdown.")



"""
NoSQL storage backend for ActivityPub data.
"""

import asyncio
from collections.abc import AsyncGenerator, AsyncIterator, Mapping
import hashlib
from logging import Logger
import math
from typing import Any, cast, override
from uuid import uuid4

from fastapi.datastructures import UploadFile
import gridfs
from pydantic import SecretStr
from pymongo.asynchronous.collection import AsyncCollection
from pymongo.asynchronous.database import AsyncDatabase
from pymongo.asynchronous.mongo_client import AsyncMongoClient

from phederation.models import APDocument, CRUD
from phederation.models.activities import APActivity, APFollow
from phederation.models.actors import APAccount, APActor
from phederation.models.collections import APOrderedCollectionPage
from phederation.models.keys import APPrivateKey
from phederation.models.maintenance import APMaintenanceCommand
from phederation.models.media import MediaMetadata
from phederation.models.objects import APCollectionItem, APObject
from phederation.models.proofs import DataIntegrityProof
from phederation.storage.nosqlcrud import NoSqlCRUD, nosql_build_search_query
from phederation.utils import ObjectId
from phederation.utils.base import (
    AccessType,
    ActivityPubBaseWithId,
    NodeInfo,
    UrlType,
    assemble_id_url,
)
from phederation.utils.exceptions import StorageError, catch_exceptions
from phederation.utils.logging import configure_logger
from phederation.utils.settings import PhedSettings

from .base import StorageBackend, fill_database_url_from_env


class NoSQLStorageBackend(StorageBackend):
    """NoSQL storage backend implementation."""

    def __init__(self, settings: PhedSettings, database_url: SecretStr):
        """Initialize storage backend."""
        super().__init__(database_url=database_url, settings=settings)
        self.logger: Logger = configure_logger(__name__, prefix=self.settings.federation.logging_prefix)

        # Create CRUD objects
        self._crud_tables: dict[str, CRUD[Any]] = {}

        self.actor: CRUD[APActor] = self.table(APActor, name=UrlType.Actors.value)
        self.account: CRUD[APAccount] = self.table(APAccount, name=UrlType.Accounts.value)
        self.activity: CRUD[APActivity] = self.table(APActivity, name=UrlType.Activities.value)
        self.object: CRUD[APObject] = self.table(APObject, name=UrlType.Objects.value)
        self.collection: CRUD[APOrderedCollectionPage] = self.table(APOrderedCollectionPage, name=UrlType.Collections.value)
        self.collection_items: CRUD[APCollectionItem] = self.table(APCollectionItem, name=UrlType.CollectionItems.value)
        self.key: CRUD[APPrivateKey] = self.table(APPrivateKey, name=UrlType.Keys.value)
        self.proof: CRUD[DataIntegrityProof] = self.table(DataIntegrityProof, name=UrlType.Proofs.value)
        self.follow: CRUD[APFollow] = self.table(APFollow, name=UrlType.Follows.value)
        self.media: CRUD[APDocument] = self.table(APDocument, name=UrlType.Media.value)
        self.node_info: CRUD[NodeInfo] = self.table(NodeInfo, name=UrlType.NodeInfo.value)
        self.maintenance: CRUD[APMaintenanceCommand] = self.table(APMaintenanceCommand, name=UrlType.Maintenance.value)

    @override
    def table[T: ActivityPubBaseWithId](self, model_class: type[T], name: str) -> CRUD[T]:
        if not name in self._crud_tables:
            self._crud_tables[name] = NoSqlCRUD[model_class](
                get_collection=self.get_nosql_collection,
                base_url=self.settings.domain.base_url,
                table_name=name,
                logger=configure_logger(f"{__name__}_{name}", prefix=self.settings.federation.logging_prefix),
                MODEL=model_class,
            )
        return self._crud_tables[name]

    async def get_nosql_collection(self, table_name: str) -> AsyncCollection[Any]:
        return (await self.database()).get_collection(table_name)

    @catch_exceptions(StorageError, "Failed to initialize nosql storage")
    @override
    async def initialize(
        self,
    ) -> None:
        """Initialize database connection."""
        self._database: AsyncDatabase[Any] | None = None
        self._client: AsyncMongoClient[Any] | None = None
        self.database_url: SecretStr = fill_database_url_from_env(self.database_url)

    async def database(self):
        loop = asyncio.get_running_loop()
        if self._client is None or self._database is None or self._client._loop != loop or self._client._closed:  # pyright: ignore[reportPrivateUsage]
            try:
                self._client = AsyncMongoClient(self.database_url.get_secret_value())
                self._database = self._client.get_database(self.settings.storage.database.name)
            except Exception as e:
                raise StorageError(f"Could not connect to database; error {type(e).__name__}, {e}")
        return self._database

    @catch_exceptions(StorageError, "NoSql: Failed to reconnect to database")
    @override
    async def reconnect(self) -> None:
        """Reconnect to the database. Useful if the async event loop changed, e.g. in delivery workers in different processes."""
        self._database = None
        _ = await self.database()

    @override
    async def paginate(self, collection_id: ObjectId, access: AccessType = AccessType.PUBLIC) -> AsyncGenerator[list[APCollectionItem], Any]:
        collection_items = await self.get_nosql_collection(await self.collection_items.get_table_name())
        query = nosql_build_search_query(collection_id=collection_id, access=access)
        total_items = await collection_items.count_documents(query)
        n_pages = max(1, math.ceil(total_items / float(self.settings.storage.page_size)))
        for page in range(0, n_pages):
            n_skip = self.settings.storage.page_size * page
            items: list[APCollectionItem] = []
            async with collection_items.find(query).skip(n_skip).limit(self.settings.storage.page_size) as cursor:
                items.extend([APCollectionItem.deserialize(item) async for item in cursor])  # pyright: ignore[reportArgumentType, reportAny]
            yield items

    @override
    async def get_collection_page_items(
        self, collection_id: ObjectId, page: int = 1, access: AccessType = AccessType.PUBLIC
    ) -> list[APCollectionItem] | None:
        collection_items = await self.get_nosql_collection(await self.collection_items.get_table_name())
        n_skip = self.settings.storage.page_size * max(0, page - 1)
        items: list[APCollectionItem] = []

        # construct the search query based on the access level
        query = nosql_build_search_query(collection_id=collection_id, access=access)

        async with collection_items.find(query).skip(n_skip).limit(self.settings.storage.page_size) as cursor:
            async for item in cursor:  # pyright: ignore[reportAny]
                items.append(APCollectionItem.deserialize(item))  # pyright: ignore[reportArgumentType, reportAny]
        return items

    @override
    async def save(self, document_id: ObjectId, content: UploadFile) -> ObjectId:
        database = await self.database()
        files_collection = database.fs.files
        bucket = gridfs.AsyncGridFSBucket(database)
        chunk_size_bytes = 1 * 1024 * 1024
        file_primary = str(uuid4()) + str(hash(document_id))

        file_id = assemble_id_url(
            type=UrlType.Media,
            base_url=self.settings.domain.base_url,
            primary=file_primary,
        )

        file_content_hash = hashlib.sha256()
        async with bucket.open_upload_stream_with_id(
            file_id=file_id,
            filename=file_primary,
            chunk_size_bytes=chunk_size_bytes,
            metadata={"contentType": "application/octet-stream", "filename": file_primary, "documentId": document_id},
        ) as grid_in:
            while True:
                chunk = await content.read(size=chunk_size_bytes)
                if not chunk:
                    break
                await grid_in.write(chunk)
                file_content_hash.update(chunk)

        # update the metadata of the file with the hex digest
        hash_digest = file_content_hash.hexdigest()
        file_data = await files_collection.find_one({"_id": file_id})
        if file_data:
            file_metadata = cast(dict[str, str], file_data.get("metadata", {}))  # pyright: ignore[reportAny]
            file_metadata["sha256"] = hash_digest
            file_metadata["documentId"] = document_id
            _ = await files_collection.update_one({"_id": file_id}, {"$set": {"metadata": file_metadata}})

        return file_id

    @override
    async def load(self, file_id: ObjectId) -> AsyncIterator[bytes]:
        database = await self.database()
        bucket = gridfs.AsyncGridFSBucket(database)

        file = await bucket.open_download_stream(file_id=file_id)
        return file

    @override
    async def metadata(self, file_id: ObjectId) -> MediaMetadata:
        database = await self.database()
        bucket = gridfs.AsyncGridFSBucket(database)

        file = await bucket.open_download_stream(file_id=file_id)
        metadata: Mapping[str, str] | None = file.metadata
        filename = file.name
        content_type: str = "application/json"
        hash_sha256: str | None = None
        if metadata and "contentType" in metadata:
            content_type = metadata.get("contentType", "")
        if metadata and "sha256" in metadata:
            hash_sha256 = metadata.get("sha256", "")
        if metadata and "documentId" in metadata:
            document_id = metadata.get("documentId", "")
        else:
            raise StorageError(f"documentId not in metadata of file {file_id}")
        return MediaMetadata(content_type=content_type, filename=filename, hash_sha256=hash_sha256, document_id=document_id)

    @override
    async def is_following(self, follower: str, following: str, check_accepted: bool = False) -> bool:
        """Check if follower 'follower' is following 'following'."""
        try:
            database = await self.database()
            collection = database.get_collection(await self.follow.get_table_name())
            if check_accepted:
                follow = await collection.find_one({"actor": follower, "object": following, "accepted": True})
            else:
                follow = await collection.find_one({"actor": follower, "object": following})
            return follow is not None

        except Exception as e:
            raise StorageError(f"NoSQLStorageBackend: Failed to check if follower '{follower}' is following '{following}': {e}")

    @override
    async def flush(self) -> None:
        """Flush storage to file, if possible."""
        pass

    @override
    async def clear(self):
        """Delete the database. Useful for testing."""
        # drop all collections at the start, to clean up
        database = await self.database()
        collection_names = await database.list_collection_names()
        for collection_name in collection_names:
            collection = database[collection_name]
            await collection.drop()

    @override
    async def close(self) -> None:
        """Close database connection."""
        pass



from typing import override

from fastapi import FastAPI

from phederation.federation import ActivityPubMiddleware, ActivityPubServer
from phederation.utils.settings import PhedSettings

from .base import BaseAPI
from .routes import BaseRoute as PhederationBaseRoute
from .routes.v1 import (
    AccountRoute,
    ActivityRoute,
    ActorRoute,
    AdminInstanceRoute,
    InstanceRoute,
    KeyRoute,
    MediaRoute,
    ObjectRoute,
    ProofRoute,
    SecurityRoute,
    WebfingerRoute,
)


class APIv1(BaseAPI):
    """Version 1 of the API.
    Sub-versions 1.1, 1.2, etc. are managed by this one.
    """

    @override
    async def initialize(self, app: FastAPI, settings: PhedSettings, middleware: ActivityPubMiddleware, server: ActivityPubServer) -> None:
        await super().initialize(app=app, settings=settings, server=server, middleware=middleware)

        self._routes: list[PhederationBaseRoute] = self.routes()
        for route in self._routes:
            await route.initialize(app=app, middleware=middleware, api=self, server=server)
            app.include_router(route.router)

    def routes(self) -> list[PhederationBaseRoute]:
        self.logger.info("Initializing API with basic routes.")
        result: list[PhederationBaseRoute] = [
            ActorRoute(),
            AccountRoute(),
            WebfingerRoute(),
            ActivityRoute(),
            ObjectRoute(),
            KeyRoute(),
            ProofRoute(),
            InstanceRoute(),
            SecurityRoute(),
            MediaRoute(),
        ]
        if self.settings.federation.admin_mode:
            self.logger.info("Adding additional admin routes.")
            result.extend([AdminInstanceRoute()])
        return result

    @override
    def version(self, major: bool = True, minor: bool = True, patch: bool = True) -> str:
        """API version string.

        The string consists of MAJOR.MINOR.PATCH version numbers. If any of the arguments is set to False, the version number is replaced by X.

        Examples:
          - version() = "1.0.0"
          - version(minor=False) = "1.X.0"
          - version(major=Fales, minor=False, patch=False) = "X.X.X"

        Args:
            major (bool, optional): Return major API version. Will only change if the entire API is changed. Defaults to True.
            minor (bool, optional): Return minor API version. Will change with milestones. Defaults to True.
            patch (bool, optional): Return patch API version. Will change on bugfixes, and should not break behavior of the API. Defaults to True.

        Returns:
            str: Version string.
        """
        v: list[str] = []
        if major:
            v.append("1")
        else:
            v.append("X")
        if minor:
            v.append("0")
        else:
            v.append("X")
        if patch:
            v.append("0")
        else:
            v.append("X")
        return ".".join(v)



"""
Handler for Reject activities.
"""

from typing import override

from phederation.models import APObject, dereference
from phederation.models.activities import APActivity
from phederation.utils import ActivityType, ObjectId
from phederation.utils.exceptions import HandlerError, ValidationError

from .base import ActivityHandler


class RejectHandler(ActivityHandler):
    """Handles Reject activities."""

    REJECTABLE_TYPES: list[str] = ["Follow"]

    @override
    async def process_outbox(self, activity: APActivity) -> APActivity:
        """Handle Reject activity."""
        actor_id = dereference(activity, key="actor")
        follow_activity_obj_or_str = activity.object

        if not follow_activity_obj_or_str:
            raise HandlerError("follow_activity_obj is None")

        # Resolve activity
        follow_activity_obj = await self.resolver.resolve_activity(follow_activity_obj_or_str)

        # Verify authorization
        if dereference(follow_activity_obj, key="object") != actor_id:
            raise HandlerError("Unauthorized: can only reject follow activities targeting self")

        follower_id = dereference(follow_activity_obj, key="actor")
        following_id = dereference(follow_activity_obj, key="object")

        if not follower_id or not following_id or not follow_activity_obj.id:
            raise HandlerError("follower_id or following_id or follow_activity_obj.id of activity is None")

        # Handle based on rejected activity type
        if follow_activity_obj.type == "Follow":
            # Check if follow request exists and hasn't been accepted yet
            is_following = await self.storage.is_following(follower=follower_id, following=following_id, check_accepted=True)

            if is_following:
                raise HandlerError("Follow request already accepted")

            await self._handle_reject_follow(
                follow_activity_id=follow_activity_obj.id, follower=follower_id, following=following_id, reason=activity.content
            )

        # Notify original actor's inbox
        if follower_id:
            _ = await self.delivery.deliver_activity(activity, recipients=[follower_id])

        self.logger.info(f"Handled Reject activity: {actor_id} -> {follow_activity_obj.id}")
        return activity

    async def _handle_reject_follow(self, follow_activity_id: ObjectId, follower: ObjectId, following: ObjectId, reason: str | None = None) -> None:
        """Handle rejecting a Follow."""
        self.logger.info(f"Rejecting follow of {follower} following {following}, reason: {reason}.")
        _ = await self.storage.follow.delete(id=follow_activity_id)

    @override
    async def validate(self, activity: APActivity) -> APActivity:
        """Validate Reject activity."""
        if activity.type != "Reject":
            raise ValidationError(f"Invalid activity type: '{activity.type}'")

        object_data = activity.object
        if not object_data:
            raise ValidationError("Reject must have an object")

        if isinstance(object_data, APObject):
            object_type = object_data.type
            if object_type not in self.REJECTABLE_TYPES:
                raise ValidationError(f"Cannot reject activity type: {object_type}")

        return activity

    @override
    async def process_inbox(self, activity: APActivity) -> None | str:
        """
        Receive a Reject in the inbox.
        This means the current actor tried to follow somebody and they rejected it, sending this back to the actor with this inbox.
        """
        actor_id = dereference(activity, key="actor")
        follow_activity_obj = activity.object

        if not follow_activity_obj:
            raise HandlerError("follow_activity_obj is None")

        # Resolve object if it's a string ID
        follow_activity_obj = await self.resolver.resolve_activity(follow_activity_obj)

        follower_id = dereference(follow_activity_obj, key="actor")
        following_id = dereference(follow_activity_obj, key="object")

        if not follower_id or not following_id or not follow_activity_obj.id:
            raise HandlerError("Could not dereference actor or object from follow activity")

        # Handle based on rejected activity type
        if follow_activity_obj.type == ActivityType.FOLLOW.value:
            # Check if follow request exists and hasn't been accepted yet
            is_following = await self.storage.is_following(follower=follower_id, following=following_id, check_accepted=True)

            if is_following:
                raise HandlerError("Follow request on local instance already stored as accepted")

            # remove the locally stored follow
            await self._handle_reject_follow(
                follow_activity_id=follow_activity_obj.id, follower=follower_id, following=following_id, reason=activity.content
            )
        else:
            raise HandlerError(f"Received a Reject activity that was not for a Follow activity (type: {follow_activity_obj.type})")

        self.logger.info(f"Handled Reject activity in inbox: {actor_id} -> {follow_activity_obj.id}")



from http import HTTPStatus
from typing import Annotated, override

from fastapi import HTTPException, Path, Security

from phederation.api.routes.base import BaseRoute
from phederation.models import APPublicKey, ErrorMessageKeyNotFound
from phederation.models.keys import ErrorMessageKeyRestricted
from phederation.security.authentication import UserInfo
from phederation.utils.base import UrlType, assemble_id_url


class KeyRoute(BaseRoute):

    @override
    def setup(self):

        @self.router.get(
            "/keys/{key_primary}",
            status_code=HTTPStatus.OK,
            response_model_exclude_none=True,
            responses={
                HTTPStatus.NOT_FOUND: {
                    "model": ErrorMessageKeyNotFound,
                    "description": "If the key with the given primary cannot be found on this instance.",
                },
            },
            tags=["keys"],
        )
        async def get_key(  # pyright: ignore[reportUnusedFunction]
            key_primary: Annotated[str, Path(description="The identifier of the key (without the URL) on this instance.")],
            user_info: Annotated[UserInfo, Security(self.middleware.authentication, scopes=["user"])],
        ) -> APPublicKey:
            """Handle requests for public keys."""
            self.logger.debug(f"Received request for key primary: {key_primary}")

            # Return key document with proper JSON-LD context
            requested_key_url = assemble_id_url(type=UrlType.Keys, base_url=self.api.settings.domain.base_url, primary=key_primary)
            key_response = await self.server.key_manager.get_public_key_APObject(key_id=requested_key_url)

            actor_requesting_access = user_info.actor_id if user_info else None
            key_with_access = await self.api.server.resolver.resolve_access_type(
                object=key_response, actor_requesting_access=actor_requesting_access
            )
            if not key_with_access or not isinstance(key_with_access, APPublicKey):
                self.logger.warning(f"Key visibility restricted, user: {user_info} cannot access it")
                raise HTTPException(
                    status_code=HTTPStatus.NOT_FOUND,
                    detail=ErrorMessageKeyRestricted().message,
                )
            if not key_response:
                self.logger.error(f"No key found with primary '{key_primary}'")
                raise HTTPException(detail=ErrorMessageKeyNotFound().message, status_code=HTTPStatus.NOT_FOUND)

            return key_response



"""API package for ActivityPub phederation.

This module implements version v1 of the Phederation API with FastAPI routes.
"""

from .base import BaseAPI
from .v1 import APIv1


async def create_api(version: str) -> BaseAPI:
    if version.startswith('1'):
        return APIv1()
    raise ValueError(f"API version '{version}' cannot be served. Must start with '1'")

API_LATEST = "1.0.0"

__all__ = [
    "BaseAPI", "create_api", "API_LATEST"
]



"""
Storage module for ActivityPub data persistence.

Features
- Abstract storage interface, one async nosql implementation for now
- Efficient implementation of specific ActivityPub features, such as followers and file storage in gridfs
"""

from .base import StorageBackend
from .nosql import NoSQLStorageBackend
from .nosqlcrud import NoSqlCRUD

__all__ = ["StorageBackend", "NoSQLStorageBackend", "NoSqlCRUD"]
